import pytest


def test_plugin_activation_forcing_via_activate(customized_suite, suite_test):

    @customized_suite.slashrc.include
    def __code__():
        slash.plugins.manager.activate('custom')

    result = customized_suite.run().session.results.global_result
    assert result.data['customized']


def test_plugin_activation_deactivate_via_commandline(customized_suite, suite_test):

    @customized_suite.slashrc.include
    def __code__():
        slash.plugins.manager.activate_later('custom')

    result = customized_suite.run(additional_args=['--without-custom']).session.results.global_result
    assert 'customized' not in result.data


def test_plugin_activation_from_configure_hook(customized_suite, suite_test):

    @customized_suite.slashrc.include
    def __code__():
        @slash.hooks.configure.register
        def configure_hook():
            slash.plugins.manager.activate_later('custom')

    result = customized_suite.run().session.results.global_result
    assert 'customized' in result.data


def test_plugin_deactivation_override_configure_hook(customized_suite, suite_test):

    @customized_suite.slashrc.include
    def __code__():
        @slash.hooks.configure.register
        def configure_hook():
            slash.plugins.manager.activate_later('custom')

    result = customized_suite.run(additional_args=['--without-custom']).session.results.global_result
    assert 'customized' not in result.data

def test_configure_hook_depends_on_configuration_cmdline(customized_suite, suite_test):

    @customized_suite.slashrc.include
    def __code__():
        slash.config.extend({
            'some_config': 1 // slash.conf.Cmdline(arg='--some-config'),
        })

        @slash.hooks.configure.register
        def configure_hook():
            assert slash.config.root.some_config == 1000

    result = customized_suite.run(additional_args=['--some-config=1000']).session.results.global_result


@pytest.fixture
def customized_suite(suite):

    @suite.slashrc.include
    def __code__():
        class CustomPlugin(slash.plugins.PluginInterface):

            def get_name(self):
                return 'custom'

            def session_start(self):
                slash.session.results.global_result.data['customized'] = True
        slash.plugins.manager.install(CustomPlugin())

    return suite



import pytest
from slash._compat import StringIO
from slash.reporting.console_reporter import TerminalWriterWrapper


def test_line_in_progress_no_newline(tw):
    tw.write('.')
    tw.write('.')
    assert tw._line == '..'


def test_line_in_progress_with_newline(tw):
    tw.write('this is \n a half line')
    assert tw._line == ' a half line'


def test_line_in_progress_with_end(tw):
    tw.write('line here\n')
    assert tw._line == ''


@pytest.fixture
def tw():
    return TerminalWriterWrapper(StringIO())



from .utils import TestCase
from .utils import run_tests_assert_success
import itertools
import os
import slash
import pytest


@pytest.mark.parametrize('parametrize', [True, False])
def test_class_name(suite, suite_test, test_type, parametrize):
    if parametrize:
        suite_test.add_parameter(num_values=3)
    summary = suite.run()
    for result in summary.get_all_results_for_test(suite_test):
        if test_type == 'method':
            assert result.test_metadata.class_name.startswith('Test')
            assert '(' not in result.test_metadata.class_name
        elif test_type == 'function':
            assert result.test_metadata.class_name is None
        else:
            raise NotImplementedError() # pragma: no cover


@pytest.mark.parametrize('parametrize', [True, False])
def test_function_name(suite, suite_test, test_type, parametrize):
    if parametrize:
        suite_test.add_parameter(num_values=3)

    summary = suite.run()
    for result in summary.get_all_results_for_test(suite_test):
        function_name = result.test_metadata.function_name
        assert function_name.startswith('test_')
        assert '.' not in result.test_metadata.function_name
        assert '(' not in result.test_metadata.function_name


def test_module_name_not_none_or_empty_string(suite):
    for result in suite.run().session.results:
        assert result.test_metadata.module_name


def test_test_index(suite):
    session = suite.run().session
    for index, result in enumerate(session.results):
        assert result.test_metadata.test_index0 == index
        assert result.test_metadata.test_index1 == index + 1


class TestMetadataTest(TestCase):

    def setUp(self):
        super(TestMetadataTest, self).setUp()
        self.root = self.get_new_path()
        self.filename = os.path.join(self.root, "testfile.py")
        with open(self.filename, "w") as f:
            f.write(_TEST_FILE_TEMPLATE)

        with slash.Session() as s:
            self.tests = slash.loader.Loader().get_runnables(self.filename)
            self.session = run_tests_assert_success(self.tests, session=s)
        self.results = list(self.session.results.iter_test_results())
        self.results.sort(key=lambda result: str(result.test_metadata))

    def test_tests_have_correct_metadata(self):
        for test, result in zip(self.tests, self.session.results.iter_test_results()):
            self.assertIs(test.__slash__, result.test_metadata)

    def test_simple_test_address(self):
        self.assertEquals(self.results[0].test_metadata.address, "{0}:T001.test_method".format(self.filename))

    def test_parameterized_test_address(self):
        parameterized = set(x.test_metadata.address for x in self.results[1:])

        self.assertEquals(parameterized, set(
            "{0}:T002.test_parameters(before:a={1}, after:c={2}, b={3})".format(self.filename, a, c, b)
            for a, b, c in itertools.product([1, 2], [3, 4], [5, 6])))

_TEST_FILE_TEMPLATE = """
import slash

class T001(slash.Test):
    def test_method(self):
        pass

class T002(slash.Test):

    @slash.parameters.iterate(a=[1, 2])
    def before(self, a):
        pass

    @slash.parameters.iterate(b=[3, 4])
    def test_parameters(self, b):
        pass

    @slash.parameters.iterate(c=[5, 6])
    def after(self, c):
        pass
"""



import pytest

from .utils.suite_writer import Suite


@pytest.mark.parametrize('num_tests', [1, 10])
def test_len(num_tests):
    s = Suite()
    for i in range(num_tests):
        s.add_test()

    assert len(s) == num_tests


def test_no_debug_info():
    s = Suite()
    s.debug_info = False

    for i in range(10):
        s.add_test()

    f = s.slashconf.add_fixture()
    s[3].depend_on_fixture(f)
    assert s.run().ok()



import functools

from ..ctx import context
from ..core.markers import repeat as repeat_marker
from ..exceptions import SkipTest


def skip_test(*args):
    """
    Skips the current test execution by raising a :class:`slash.exceptions.SkipTest`
    exception. It can optionally receive a reason argument.
    """
    raise SkipTest(*args)

def repeat(num_repetitions):
    """
    Marks a test to be repeated multiple times when run
    """
    return repeat_marker(num_repetitions)


def skipped(thing, reason=None):
    """
    A decorator for skipping methods and classes
    """
    from ..core.test import Test

    if isinstance(thing, str):
        return functools.partial(skipped, reason=thing)
    if isinstance(thing, type) and issubclass(thing, Test):
        thing.skip_all(reason)
        return thing

    @functools.wraps(thing)
    def new_func(*_, **__):  # pylint: disable=unused-argument
        skip_test(reason)
    return new_func


def add_error(msg=None, frame_correction=0):
    """
    Adds an error to the current test result

    :param msg: can be either an object or a string representing a message
    :param frame_correction: when delegating add_error from another function, specifies
      the amount of frames to skip to reach the actual cause of the added error
    """
    if context.session is not None:
        return context.session.results.current.add_error(msg, frame_correction=frame_correction+1)


def add_failure(msg=None, frame_correction=0):
    """
    Adds a failure to the current test result

    :param msg: can be either an object or a string representing a message
    :param frame_correction: when delegating add_failure from another function, specifies
      the amount of frames to skip to reach the actual cause of the added failure
    """
    if context.session is not None:
        return context.session.results.current.add_failure(msg, frame_correction=frame_correction+1)


def set_test_detail(key, value):
    """
    Store an object providing additional information about the current running test in a certain key.
    Each test has its own storage.

    :param key: a hashable object
    :param value: can be either an object or a string representing additional details
    """
    if context.session is not None:
        context.session.results.current.set_test_detail(key, value)



import logbook

import pytest
import slash
from slash._compat import StringIO
from slash.log import ConsoleHandler


@pytest.mark.parametrize('use_truncation', [True, False])
def test_line_truncation(long_text, use_truncation, config_override):

    config_override('log.truncate_console_lines', use_truncation)
    config_override('log.console_level', logbook.DEBUG)

    console = StringIO()

    with slash.Session(console_stream=console):
        logbook.debug(long_text)

    if use_truncation:
        assert long_text not in console.getvalue()
        assert long_text[:20] in console.getvalue()
    else:
        assert long_text in console.getvalue()


@pytest.fixture
def long_text():
    return 'a' * 200



import itertools
from contextlib import contextmanager

from .test import Test


class MethodTest(Test):

    def __init__(self, suite, cls):
        super(MethodTest, self).__init__(suite, cls.file)
        self.cls = cls

    def is_method_test(self):
        return True

    def _get_argument_names(self):
        return itertools.chain(['self'], super(MethodTest, self)._get_argument_names())



import pytest
import gossip
import slash
import slash.runner
from slash import exception_handling, Session
from slash.loader import Loader


@pytest.mark.parametrize('other_error', ['error', 'failure', None])
def test_success_only_cleanups_with_skips(suite, suite_test, other_error):


    if other_error == 'error':
        suite_test.append_line('slash.add_error("error")')
        suite_test.expect_error()
    elif other_error == 'failure':
        suite_test.append_line('slash.add_failure("failure")')
        suite_test.expect_failure()
    elif other_error is not None:
        raise NotImplementedError() # pragma: no cover

    @suite_test.append_body
    def __code__():
        def callback():
            __ut__.events.add('success_only_cleanup_called')
        slash.add_cleanup(callback, success_only=True)
        slash.skip_test()


    if other_error is None:
        suite_test.expect_skip()

    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)
    assert result.has_skips()
    assert ('success_only_cleanup_called' in summary.events) == (other_error is None)



def test_fatal_exceptions_from_cleanup(suite, suite_test, is_last_test):

    @suite_test.append_body
    def __code__():
        @slash.add_cleanup
        def cleanup():
            from slash.exception_handling import mark_exception_fatal
            raise mark_exception_fatal(Exception())

    suite_test.expect_error()

    for t in suite.iter_all_after(suite_test, assert_has_more=not is_last_test):
        t.expect_not_run()

    suite.run()


def test_add_skip_from_test_cleanup(suite, suite_test):
    cleanup = suite_test.add_deferred_event(decorator='slash.add_cleanup', extra_code=['slash.skip_test()'])
    suite_test.expect_skip()
    summary = suite.run()
    assert summary.events[cleanup].timestamp


@pytest.mark.parametrize('cleanup_mechanism', ['this', 'slash'])
def test_add_skip_from_fixture_cleanup(suite, suite_test, cleanup_mechanism):
    suite_test.expect_skip()
    fixture = suite.slashconf.add_fixture()
    suite_test.depend_on_fixture(fixture)
    cleanup = fixture.add_deferred_event(decorator='{0}.add_cleanup'.format(cleanup_mechanism), extra_code=['slash.skip_test()'])
    summary = suite.run()
    assert summary.events[cleanup].timestamp



def test_test_cleanups_happen_before_fixture_cleanups(suite, suite_test):
    fixture = suite.slashconf.add_fixture()
    suite_test.depend_on_fixture(fixture)
    fixture_cleanup = fixture.add_cleanup()
    test_cleanup = suite_test.add_cleanup()
    events = suite.run().events

    assert events[fixture_cleanup].timestamp > events[test_cleanup].timestamp


def test_cannot_add_cleanup_without_active_session():
    with pytest.raises(RuntimeError):
        slash.add_cleanup(lambda: None)


def test_cleanups(suite, suite_test):

    cleanup = suite_test.add_cleanup()

    summary = suite.run()

    assert summary.events[cleanup]


def test_cleanup_args_kwargs_deprecated():
    with slash.Session() as s:
        slash.add_cleanup(lambda: None, "arg1", arg2=1)
    [w] = s.warnings
    assert 'deprecated' in str(w).lower()


def test_cleanup_ordering(suite, suite_test):
    cleanup1 = suite_test.add_cleanup()
    cleanup2 = suite_test.add_cleanup()
    events = suite.run().events

    assert events[cleanup1].timestamp > events[cleanup2].timestamp


@pytest.mark.parametrize('fail_test', [True, False])
def test_errors_in_cleanup(suite, suite_test, fail_test):
    cleanup1 = suite_test.add_cleanup()
    cleanup2 = suite_test.add_cleanup(extra_code=['None.a = 2'])

    if fail_test:
        suite_test.when_run.raise_exception()
    else:
        suite_test.expect_error()

    summary = suite.run()

    assert summary.events[cleanup1].timestamp > summary.events[cleanup2].timestamp

    [result] = summary.get_all_results_for_test(suite_test)

    assert len(result.get_errors()) == 2 if fail_test else 1
    cleanup_error = result.get_errors()[-1]
    assert 'AttributeError' in str(cleanup_error)
    assert 'NoneType' in str(cleanup_error)





#!/usr/bin/env python
from __future__ import print_function
import argparse
import contextlib
import logbook # pylint: disable=F0401
import sys

from ..__version__ import __version__

_COMMANDS = {
    "run": "slash.frontend.slash_run:slash_run",
    "resume": "slash.frontend.slash_run:slash_resume",
    "version": "slash.frontend.main:slash_version",
    "list": "slash.frontend.slash_list:slash_list",
    "list-config": "slash.frontend.list_config:list_config",
}


def _get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="Available commands:\n\t{0}".format("\n\t".join(sorted(_COMMANDS))),
        usage="%(prog)s command...",
    )

    parser.add_argument("-v", action="append_const", const=1, dest="verbosity", default=[],
                        help="Be more verbose. Can be specified multiple times to increase verbosity further")
    parser.add_argument("cmd")
    parser.add_argument("argv", nargs=argparse.REMAINDER)
    return parser

def main():
    parser = _get_parser()
    args = parser.parse_args()
    with _setup_logging_context(args):
        module_name = _COMMANDS.get(args.cmd)
        if not module_name:
            parser.error("No such command: {0}".format(args.cmd))
        module_name, func_name = module_name.split(":")
        module = __import__(module_name, fromlist=[""])
        func = getattr(module, func_name)
        return func(args.argv)
    return 0

def slash_version(_):
    print('Slash v{0}'.format(__version__))

################################## Boilerplate ################################
_DEFAULT_LOG_LEVEL = logbook.WARNING
@contextlib.contextmanager
def _setup_logging_context(args):
    log_level = max(logbook.DEBUG, _DEFAULT_LOG_LEVEL - len(args.verbosity))
    with logbook.NullHandler().applicationbound():
        with logbook.StderrHandler(level=log_level, bubble=False).applicationbound():
            yield

#### For use with entry_points/console_scripts
def main_entry_point():
    sys.exit(main())

if __name__ == "__main__":
    main_entry_point()



import argparse
import os
import sys
from contextlib import contextmanager

import colorama

from .. import conf, plugins, hooks
from .._compat import cStringIO, iteritems, itervalues
from ..plugins import UnknownPlugin


@contextmanager
def get_cli_environment_context(argv=None, config=conf.config, extra_args=(), positionals_metavar=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = SlashArgumentParser(prog=_deduce_program_name(), positionals_metavar=positionals_metavar)
    if extra_args:
        _populate_extra_args(parser, extra_args)
    _configure_parser_by_plugins(parser)
    _configure_parser_by_config(parser, config)

    try:
        argv = _add_pending_plugins_from_commandline(argv)
    except UnknownPlugin as e:
        parser.error(str(e))

    if positionals_metavar is not None:
        parsed_args, positionals = parser.parse_known_args(argv)
    else:
        parsed_args = parser.parse_args(argv)
        positionals = []

    with _get_modified_configuration_from_args_context(parser, config, parsed_args):
        hooks.configure() # pylint: disable=no-member
        plugins.manager.activate_pending_plugins()
        parsed_args.positionals = positionals
        _configure_plugins_from_args(parsed_args)
        yield parser, parsed_args

def _deduce_program_name():
    returned = os.path.basename(sys.argv[0])
    if len(sys.argv) > 1:
        returned += " {0}".format(sys.argv[1])
    return returned

def _populate_extra_args(parser, extra_args):
    for argument in extra_args:
        parser.add_argument(*argument.args, **argument.kwargs)

_PLUGIN_ACTIVATION_PREFIX = "--with-"
_PLUGIN_DEACTIVATION_PREFIX = "--without-"
def _add_pending_plugins_from_commandline(argv):
    returned_argv = []
    for arg in argv:
        if arg.startswith(_PLUGIN_DEACTIVATION_PREFIX):
            plugin_name = arg[len(_PLUGIN_DEACTIVATION_PREFIX):]
            plugins.manager.deactivate_later(plugin_name)
        elif arg.startswith(_PLUGIN_ACTIVATION_PREFIX):
            plugin_name = arg[len(_PLUGIN_ACTIVATION_PREFIX):]
            plugins.manager.activate_later(plugin_name)
        else:
            returned_argv.append(arg)
    return returned_argv

def _configure_parser_by_plugins(parser):
    for plugin in itervalues(plugins.manager.get_installed_plugins()):
        group = parser.add_argument_group('Options for --with-{0}'.format(plugin.get_name()))
        plugin.configure_argument_parser(group)

def _configure_plugins_from_args(args):
    for plugin in itervalues(plugins.manager.get_active_plugins()):
        plugin.configure_from_parsed_args(args)

def _configure_parser_by_config(parser, config):
    parser.add_argument(
        "-o", dest="config_overrides", metavar="PATH=VALUE", action="append",
        default=[],
        help="Provide overrides for configuration"
    )
    for path, node, cmdline in _iter_cmdline_config(config):
        cmdline.configure_parser(parser, path, node)

def _iter_cmdline_config(config):
    for path, cfg in config.traverse_leaves():
        cmdline = (cfg.metadata or {}).get("cmdline")
        if cmdline is None:
            continue
        yield path, cfg, cmdline

@contextmanager
def _get_modified_configuration_from_args_context(parser, config, args):
    to_restore = []
    try:
        for path, cfg, cmdline in _iter_cmdline_config(config):
            old_value = cfg.get_value()
            new_value = cmdline.update_value(old_value, args)
            if new_value != old_value:
                to_restore.append((path, cfg.get_value()))
                config.assign_path(path, new_value, deduce_type=True, default_type=str)
        for override in args.config_overrides:
            if "=" not in override:
                parser.error("Invalid config override: {0}".format(override))
            path, _ = override.split("=", 1)
            to_restore.append((path, config.get_path(path)))
            try:
                config.assign_path_expression(override, deduce_type=True, default_type=str)
            except ValueError:
                parser.error("Invalid value for config override: {0}".format(override))
        yield
    finally:
        for path, prev_value in reversed(to_restore):
            config.assign_path(path, prev_value)

class SlashArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        positionals_metavar = kwargs.pop("positionals_metavar", None)
        super(SlashArgumentParser, self).__init__(*args, **kwargs)
        self._positionals_metavar = positionals_metavar


    def format_help(self):
        returned = cStringIO()
        helpstring = super(SlashArgumentParser, self).format_help()
        helpstring = self._tweak_usage_positional_metavars(helpstring)
        returned.write(helpstring)
        return returned.getvalue()

    def _tweak_usage_positional_metavars(self, usage):
        """
        Adds fake positionals metavar at the end of the usage line
        """
        if self._positionals_metavar is None:
            return usage

        returned = ""
        added_metavars = False
        for line in cStringIO(usage):
            if not added_metavars and not line.strip():
                returned = returned[:-1] + " [{0} [{0} ...]] \n".format(self._positionals_metavar)
            returned += line
        return returned

    def _iter_available_plugins(self):
        active_plugin_names = set(plugins.manager.get_active_plugins())
        for plugin_name, plugin in iteritems(plugins.manager.get_installed_plugins()):
            if plugin_name not in active_plugin_names:
                yield plugin_name, plugin

class Argument(object):
    """
    helper to defer initialization of cmdline parsers to later stages
    """
    def __init__(self, *args, **kwargs):
        super(Argument, self).__init__()
        self.args = args
        self.kwargs = kwargs

COLOR_RESET = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL  # pylint: disable=no-member


def make_styler(style):
    return lambda s: '{0}{1}{2}'.format(style, s, COLOR_RESET)

UNDERLINED = '\x1b[4m'



# pylint: disable-msg=W0201
import functools

import pytest
import slash
from slash.exceptions import NoActiveSession
from slash.runner import run_tests


@pytest.mark.parametrize('adder', ['add_failure', 'add_error'])
def test_stop_on_error_with_error_and_skip(suite, adder):
    suite_test = suite[2]
    suite_test.append_line('slash.{0}("err")'.format(adder))
    suite_test.append_line('slash.skip_test()')
    suite_test.expect_skip()

    for test in suite.iter_all_after(suite_test, assert_has_more=True):
        test.expect_not_run()

    summary = suite.run(additional_args=['-x'])
    [result] = summary.get_all_results_for_test(suite_test)

    assert result.has_skips()
    assert result.has_errors_or_failures()


def test_run_tests_fails_without_active_session():
    with pytest.raises(NoActiveSession):
        run_tests([])


def test_simple_run(suite):
    suite.run()


def test_iter_results_ordering(suite):
    for index, test in enumerate(suite):
        test.append_line('slash.context.result.data["index"] = {0}'.format(index))

    results = list(suite.run().session.results.iter_test_results())

    for index, result in enumerate(results):
        assert result.data['index'] == index


def test_failed(suite, suite_test):
    suite_test.when_run.fail()
    result = suite.run()[suite_test]
    assert result.is_failure()
    assert not result.is_error()
    assert not result.is_success()


def test_error(suite, suite_test):
    suite_test.when_run.error()
    result = suite.run()[suite_test]
    assert result.is_error()
    assert not result.is_failure()
    assert not result.is_success()


def test_skip(suite, suite_test):
    suite_test.when_run.skip()
    result = suite.run()[suite_test]
    assert result.is_skip()
    assert not result.is_error()
    assert not result.is_failure()
    assert not result.is_success()


def test_stop_on_fatal_exception(suite, suite_test, fatal_error_adder):
    fatal_error_adder(suite_test)
    for remaining_test in suite.iter_all_after(suite_test):
        remaining_test.expect_not_run()

    suite.run()


def test_stop_on_error(suite, suite_test, failure_type):
    if failure_type == 'error':
        suite_test.when_run.error()
    elif failure_type == 'failure':
        suite_test.when_run.fail()
    else:
        raise NotImplementedError()  # pragma: no cover

    for test in suite.iter_all_after(suite_test):
        test.expect_not_run()

    suite.run(additional_args=['-x'])


def test_stop_on_error_unaffected_by_skips(suite, suite_test):
    suite_test.when_run.skip()
    summary = suite.run(additional_args=['-x'])
    for test in suite.iter_all_after(suite_test):
        for res in summary.get_all_results_for_test(test):
            assert res.is_success()



def test_debug_if_needed(request, config_override, suite, suite_test):
    suite_test.when_run.fail()

    debugged = {'value': False}

    def _debug_if_needed(exc_info):
        debugged['value'] = True

    request.addfinalizer(functools.partial(
        setattr, slash.utils.debug, '_KNOWN_DEBUGGERS', slash.utils.debug._KNOWN_DEBUGGERS))
    slash.utils.debug._KNOWN_DEBUGGERS = [_debug_if_needed]

    config_override('debug.enabled', True)

    suite.run()

    assert debugged['value']

@pytest.fixture(params=['raising', 'adding'])
def fatal_error_adder(request, failure_type):

    def adder(test):
        if request.param == 'raising':
            test.append_line(
                'from slash.exception_handling import mark_exception_fatal')
            test.append_line('raise mark_exception_fatal({0}())'.format('AssertionError' if failure_type == 'failure' else 'Exception'))
        elif request.param == 'adding':
            test.append_line('slash.add_{0}("msg").mark_fatal()'.format(failure_type))
        else:
            raise NotImplementedError() # pragma: no cover

        getattr(test, 'expect_{0}'.format(failure_type))()
    return adder


@pytest.fixture(params=['failure', 'error'])
def failure_type(request):
    return request.param



from uuid import uuid4
import itertools
from contextlib import contextmanager

from .code_element import CodeElement
from .element import Element
from .fixture import Fixture
from .nonmethod_test import NonMethodTest
from .test_class import Class
from .test_container import TestContainer


class File(TestContainer, CodeElement):

    def __init__(self, suite, relpath=None):
        super(File, self).__init__(suite)
        self._classes = []
        self._fixtures = []

        if relpath is None:
            relpath = 'test_{0}.py'.format(self.id)
        self._relpath = relpath

    def add_hook_event(self, hook_name, extra_args=(), evt_name='evt'):
        self.append_line('@slash.hooks.{0}.register'.format(hook_name))
        event_code = '{0}_{1}'.format(evt_name, uuid4())
        self.append_line('def _hook():')
        self.append_line('    __ut__.events.add({0!r}, {1})'.format(
            event_code, ', '.join(extra_args)))
        return event_code

    @property
    def classes(self):
        return list(self._classes)

    def add_fixture(self, **kw):
        returned = Fixture(self.suite, self, **kw)
        self._fixtures.append(returned)
        return returned

    def get_relative_path(self):
        return self._relpath

    def add_class(self):
        cls = Class(self.suite, self)
        self._classes.append(cls)
        return cls

    def get_last_class(self):
        if not self._classes:
            return None
        return self._classes[-1]

    def add_function_test(self):
        returned = NonMethodTest(self.suite, self)
        self._tests.append(returned)
        self.suite.notify_test_added(returned)
        return returned

    @contextmanager
    def _body_context(self, code_formatter):
        with super(File, self)._body_context(code_formatter):
            if self.suite.debug_info:
                code_formatter.writeln('import __ut__')
            code_formatter.writeln('import slash')
            code_formatter.writeln()
            yield None

    def _write_body(self, code_formatter):
        super(File, self)._write_body(code_formatter)
        for thing in itertools.chain(self._classes, self._tests, self._fixtures):
            thing.write(code_formatter)



import re
import pytest


def test_cyclic_fixture_detection(suite, suite_test):
    fixture1 = suite.slashconf.add_fixture()
    fixture2 = suite.slashconf.add_fixture()
    fixture1.depend_on_fixture(fixture2)
    fixture2.depend_on_fixture(fixture1)

    suite_test.depend_on_fixture(fixture1)
    for test in suite:
        test.expect_deselect()

    summary = suite.run()
    assert not summary.session.results.global_result.is_success()
    assert re.search(r'yclic fixture dependency detected in \S+: {0} -> {1} -> {0}'.format(fixture1.name, fixture2.name),
                     summary.get_console_output())


def test_cyclic_fixture_detection_depend_on_self(suite, suite_test):
    fixture = suite.slashconf.add_fixture()
    fixture.depend_on_fixture(fixture)

    suite_test.depend_on_fixture(fixture)
    for test in suite:
        test.expect_deselect()

    summary = suite.run()
    assert not summary.session.results.global_result.is_success()
    assert re.search(r'yclic fixture dependency detected in \S+: {0} depends on itself'.format(fixture.name),
                     summary.get_console_output())



import re

import pytest

from slash import plugins
from slash import hooks
from slash.plugins import PluginInterface, IncompatiblePlugin, UnknownPlugin
from .utils import CustomException


def test_get_active_plugins(plugin):
    plugins.manager.install(plugin)
    assert plugins.manager.get_active_plugins() == {}
    plugins.manager.activate(plugin)
    assert plugins.manager.get_active_plugins() == {plugin.get_name(): plugin}
    plugins.manager.deactivate(plugin)
    assert plugins.manager.get_active_plugins() == {}


def test_deactivaion_no_activation(plugin):
    plugins.manager.install(plugin)
    assert (not plugin._deactivate_called)
    plugins.manager.uninstall(plugin)
    assert not plugin._deactivate_called, 'Deactivate called even though plugin not activated'


def test_activation_exception(plugin):
    plugin.activate = CustomException.do_raise
    plugins.manager.install(plugin)

    with pytest.raises(CustomException):
        plugins.manager.activate(plugin)

    # make sure no registrations are in effect...
    _assert_hooks_not_registered(plugin)

    plugins.manager.deactivate(plugin)
    assert (not plugin._deactivate_called), 'Deactivate unexpectedly called!'


def test_deactivation_exception(plugin):
    plugin.deactivate = CustomException.do_raise
    plugins.manager.install(plugin, activate=True)

    with pytest.raises(CustomException):
        plugins.manager.deactivate(plugin)
    _assert_hooks_not_registered(plugin)


def test_activate_called(plugin):
    plugins.manager.install(plugin)
    assert (not plugin._activate_called)
    plugins.manager.activate(plugin)
    assert plugin._activate_called


def test_deactivate_called_on_deactivate(plugin):
    plugins.manager.install(plugin)
    assert (not plugin._deactivate_called)
    plugins.manager.activate(plugin)
    assert (not plugin._deactivate_called)
    plugins.manager.deactivate(plugin)
    assert plugin._deactivate_called


def test_hook_registration(plugin):
    plugins.manager.install(plugin)
    _assert_hooks_not_registered(plugin)
    plugins.manager.activate(plugin)
    hooks.session_start()
    assert plugin.session_start_call_count == 1
    plugins.manager.deactivate(plugin)
    hooks.session_start()
    assert plugin.session_start_call_count == 1


def test_install_and_activate(plugin):
    """test plugins.manager.install(..., activate=True)"""
    plugins.manager.install(plugin, activate=True)
    assert plugin.get_name() in plugins.manager.get_active_plugins()


def test_uninstall_also_deactivates(plugin):
    plugins.manager.install(plugin)
    plugins.manager.activate(plugin)
    plugins.manager.uninstall(plugin)
    hooks.session_start()
    assert plugin.session_start_call_count == 0


def test_cannot_activate_uninstalled_plugin():
    class Plugin(PluginInterface):

        def get_name(self):
            return "Test plugin"
    with pytest.raises(ValueError) as caught:
        plugins.manager.activate(Plugin())
    assert re.search(r".*Unknown plugin: Test plugin.*", str(caught.value))


def test_unknown_hook_names(request):
    "Make sure that plugins with unknown hook names get discarded"
    class Plugin(PluginInterface):

        def get_name(self):
            return "Test plugin"

        def unknown_hook_1(self):
            pass

    plugin = Plugin()
    plugins.manager.install(plugin)

    @request.addfinalizer
    def cleanup():
        plugins.manager.uninstall(plugin)

    with pytest.raises(IncompatiblePlugin) as caught:
        plugins.manager.activate(plugin)

    assert re.search(r"\bUnknown hooks\b.*", str(caught.value))


def test_custom_hook_names(request):
    "Make sure that plugins with unknown hook names get discarded"
    class Plugin(PluginInterface):

        def get_name(self):
            return "Test plugin"

        def custom_hook(self):
            pass

    hooks.add_custom_hook("custom_hook")

    @request.addfinalizer
    def cleanup():
        hooks.remove_custom_hook("custom_hook")
    plugin = Plugin()
    plugins.manager.install(plugin, activate=True)
    plugins.manager.uninstall(plugin)


def test_pending_activation(plugin):
    plugins.manager.install(plugin)
    assert not plugins.manager.get_active_plugins()
    plugins.manager.activate_later(plugin)
    assert not plugins.manager.get_active_plugins()
    plugins.manager.activate_pending_plugins()
    assert plugin.get_name() in plugins.manager.get_active_plugins()
    assert plugin._activate_called


def test_pending_activation_deactivation(plugin):
    plugins.manager.install(plugin)
    plugins.manager.activate_later(plugin)
    plugins.manager.deactivate_later(plugin)
    assert plugin.get_name() in plugins.manager._pending_activation
    assert plugin.get_name() in plugins.manager._pending_deactivation
    plugins.manager.activate_pending_plugins()
    assert not plugin._activate_called


def test_install_activate_later(plugin):
    plugins.manager.install(plugin, activate_later=True)
    assert plugin.get_name() in plugins.manager._pending_activation


@pytest.mark.parametrize('activate_later_first', [True, False])
def test_deactivate_later_already_activated(plugin, activate_later_first):
    plugins.manager.install(plugin, activate=True)
    if activate_later_first:
        plugins.manager.activate_later(plugin)
    plugins.manager.deactivate_later(plugin)
    plugins.manager.activate_pending_plugins()
    assert plugin.get_name() not in plugins.manager.get_active_plugins()
    assert plugin._activate_called
    assert plugin._deactivate_called

def test_pending_activation_not_exists(plugin):
    with pytest.raises(UnknownPlugin):
        plugins.manager.activate_later(plugin)
    with pytest.raises(UnknownPlugin):
        plugins.manager.activate_later(plugin.get_name())



def test_pending_deactivation_not_exists(plugin):
    with pytest.raises(UnknownPlugin):
        plugins.manager.deactivate_later(plugin)
    with pytest.raises(UnknownPlugin):
        plugins.manager.deactivate_later(plugin.get_name())



def _assert_hooks_not_registered(plugin):
    hooks.session_start()
    assert plugin.session_start_call_count == 0, 'Hook unexpectedly registered!'




import weakref

class GarbageCollectionMarker(object):

    def mark(self, obj):
        marker = _Marker(obj)
        obj.__marker__ = _MarkerAnchor(marker)
        return marker


class _Marker(object):

    destroyed = False

    def __init__(self, obj):
        super(_Marker, self).__init__()
        self._weakref = weakref.proxy(obj)


class _MarkerAnchor(object):

    def __init__(self, marker):
        super(_MarkerAnchor, self).__init__()
        self._marker = marker

    def __del__(self):
        self._marker.destroyed = True



import itertools
import os
import random
import shutil
import tempfile

from forge import Forge

import gossip
import pytest
import slash
import slash.plugins
from slash import resuming
from slash.loader import Loader

from .utils.cartesian import Cartesian
from .utils.suite_writer import Suite
from .utils.garbage_collection import GarbageCollectionMarker


@pytest.fixture(scope='session', autouse=True)
def unittest_mode_logging():
    slash.config.root.log.unittest_mode = True


@pytest.fixture(scope='session', autouse=True)
def random_seed():
    random.seed(0xdeadface)


@pytest.fixture(scope='session', autouse=True)
def no_user_config(request):
    tmpdir = tempfile.mkdtemp()
    slash.conf.config.root.run.user_customization_file_path = os.path.join(
        tmpdir, 'slashrc')

    @request.addfinalizer
    def cleanup():
        os.rmdir(tmpdir)

@pytest.fixture
def no_plugins(request):
    slash.plugins.manager.uninstall_all()
    @request.addfinalizer
    def cleanup():
        slash.plugins.manager.uninstall_all()
        slash.plugins.manager.install_builtin_plugins()


@pytest.fixture
def forge(request):

    returned = Forge()

    @request.addfinalizer
    def cleanup():
        returned.verify()
        returned.restore_all_replacements()

    return returned


@pytest.fixture
def config_override(request):

    def _override(path, value):
        prev_value = slash.config.get_config(path).get_value()

        @request.addfinalizer
        def restore():
            slash.config.assign_path(path, prev_value)
        slash.config.assign_path(path, value)
    return _override


@pytest.fixture
def cartesian():
    return Cartesian()


@pytest.fixture(scope="function", autouse=True)
def cleanup_hook_registrations(request):
    @request.addfinalizer
    def _cleanup():
        for hook in gossip.get_group("slash").get_hooks():
            hook.unregister_all()
        assert not gossip.get_group("slash").get_subgroups()


@pytest.fixture
def checkpoint():
    return Checkpoint()

@pytest.fixture
def checkpoint1():
    return Checkpoint()

@pytest.fixture
def checkpoint2():
    return Checkpoint()


_timestamp = itertools.count(1000000)


class Checkpoint(object):

    called_count = 0
    args = kwargs = timestamp = None

    def __call__(self, *args, **kwargs):
        self.called_count += 1
        self.args = args
        self.kwargs = kwargs
        self.timestamp = next(_timestamp)

    @property
    def called(self):
        return self.called_count > 0


@pytest.fixture(autouse=True, scope="function")
def fix_resume_path(request):
    prev = resuming._RESUME_DIR
    resuming._RESUME_DIR = tempfile.mkdtemp()

    @request.addfinalizer
    def cleanup():
        shutil.rmtree(resuming._RESUME_DIR)
        resuming._RESUME_DIR = prev


@pytest.fixture
def suite_test(suite, test_type, is_last_test):
    returned = suite.add_test(type=test_type)
    if not is_last_test:
        _ = suite.add_test(type=test_type)

    return returned


@pytest.fixture(params=[True, False])
def is_last_test(request):
    return request.param


@pytest.fixture(params=['method', 'function'])
def test_type(request):
    return request.param


@pytest.fixture
def suite():
    returned = Suite()
    returned.populate()
    return returned


@pytest.fixture
def slash_session():
    return slash.Session()

@pytest.fixture
def test_loader():
    return Loader()


@pytest.fixture
def active_slash_session(request):
    returned = slash.Session()
    returned.__enter__()

    @request.addfinalizer
    def finalize():
        returned.__exit__(None, None, None)

    return returned


@pytest.fixture(params=["slashconf", "module"])
def defined_fixture(request, suite, suite_test):
    if request.param == 'slashconf':
        return suite.slashconf.add_fixture()
    elif request.param == 'module':
        return suite_test.file.add_fixture()

    raise NotImplementedError()  # pragma: no cover


@pytest.fixture
def gc_marker():
    return GarbageCollectionMarker()


@pytest.fixture(autouse=True, scope="function")
def reset_gossip(request):
    @request.addfinalizer
    def cleanup():
        for group in list(gossip.get_groups()):
            if group.name == 'slash':
                continue
            group.undefine()

        for hook in gossip.get_all_hooks():
            if hook.group.name != 'slash':
                hook.undefine()
            else:
                hook.unregister_all()


@pytest.fixture
def plugin(no_plugins):

    class StartSessionPlugin(slash.plugins.PluginInterface):
        _activate_called = False
        _deactivate_called = False

        def __init__(self):
            super(StartSessionPlugin, self).__init__()
            self.session_start_call_count = 0

        def get_name(self):
            return "start-session"

        def session_start(self):
            self.session_start_call_count += 1

        def activate(self):
            self._activate_called = True

        def deactivate(self):
            self._deactivate_called = True
    return StartSessionPlugin()



import gc


def test_gc_marker(gc_marker):

    class Obj(object):
        pass

    obj = Obj()
    marker = gc_marker.mark(obj)
    assert not marker.destroyed
    del obj
    gc.collect()
    assert marker.destroyed



import itertools
import functools
import os
import sys
from contextlib import contextmanager

import logbook

from ..app import get_application_context
from ..conf import config
from ..exceptions import SlashException
from ..exception_handling import handling_exceptions
from ..resuming import (get_last_resumeable_session_id, get_tests_to_resume,
                        save_resume_state)
from ..runner import run_tests
from ..utils.interactive import generate_interactive_test

_logger = logbook.Logger(__name__)

def slash_run(args, report_stream=None, resume=False, app_callback=None, test_sort_key=None):
    if report_stream is None:
        report_stream = sys.stderr
    with _get_slash_app_context(args, report_stream, resume) as app:
        if app_callback is not None:
            app_callback(app)
        try:
            with handling_exceptions():
                if resume:
                    session_ids = app.args.positionals
                    if not session_ids:
                        session_ids = [get_last_resumeable_session_id()]
                    to_resume = [x for session_id in session_ids for x in get_tests_to_resume(session_id)]
                    collected = app.test_loader.get_runnables(to_resume, sort_key=test_sort_key)
                else:
                    collected = _collect_tests(app, args, test_sort_key=test_sort_key)
                if app.args.interactive:
                    collected = itertools.chain([generate_interactive_test()], collected)
            with app.session.get_started_context():
                run_tests(collected)

        except SlashException as e:
            logbook.error(str(e))
            return -1
        finally:
            save_resume_state(app.session.results)

        if app.session.results.is_success(allow_skips=True):
            return 0
        return -1

@contextmanager
def _get_slash_app_context(args, report_stream, resume_session):
    with get_application_context(
            argv=args, positionals_metavar="SESSION_ID" if resume_session else "TEST",
            enable_interactive=True,
            report_stream=report_stream) as app:
        yield app

slash_resume = functools.partial(slash_run, resume=True)

def _collect_tests(app, args, test_sort_key=None):  # pylint: disable=unused-argument
    paths = app.args.positionals

    paths = _extend_paths_from_suite_files(paths)

    if not paths and not app.args.interactive:
        paths = config.root.run.default_sources


    if not paths and not app.args.interactive:
        app.error("No tests specified")

    collected = app.test_loader.get_runnables(paths, sort_key=test_sort_key)
    if len(collected) == 0 and not app.args.interactive:
        app.error("No tests could be collected", usage=False)

    return collected

def _extend_paths_from_suite_files(paths):
    suite_files = config.root.run.suite_files
    if not suite_files:
        return paths
    paths = list(paths)
    paths.extend(_iter_suite_file_paths(suite_files))
    return paths

def _iter_suite_file_paths(suite_files):
    for filename in suite_files:

        dirname = os.path.abspath(os.path.dirname(filename))
        for path in open(filename):
            path = path.strip()
            if not path or path.startswith("#"):
                continue

            if not os.path.isabs(path):
                path = os.path.abspath(os.path.join(dirname, path))

            if not path.endswith('.py') and '.py:' not in path and not os.path.isdir(path):
                for p in _iter_suite_file_paths([path]):
                    yield p
                continue

            yield path



import os

import gossip
import pytest
import slash
from slash._compat import PY2
from slash import hooks, plugins
from slash.plugins import IncompatiblePlugin, PluginInterface

from .utils import CustomException, TestCase


@pytest.fixture
def restore_plugins_on_cleanup(request):
    request.addfinalizer(plugins.manager.install_builtin_plugins)
    request.addfinalizer(plugins.manager.uninstall_all)


def test_registers_on_none(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on(None)
        def some_method_here(self):
            checkpoint()

    gossip.trigger('slash.some_method_here')
    assert not checkpoint.called



def test_registers_on_with_private_methods(restore_plugins_on_cleanup, checkpoint):

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

        @plugins.registers_on('some_hook')
        def _handler(self):
            checkpoint()

    assert not checkpoint.called
    gossip.trigger('some_hook')
    assert checkpoint.called


def test_class_variables_allowed(restore_plugins_on_cleanup):
    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        ATTRIBUTE = 'some_value'

        def get_name(self):
            return 'sample'


def test_active_decorator(restore_plugins_on_cleanup):

    plugins.manager.uninstall_all()

    @slash.plugins.active
    class SamplePlugin(PluginInterface):

        def get_name(self):
            return 'sample'

    assert isinstance(SamplePlugin, type)
    assert issubclass(SamplePlugin, PluginInterface)
    [active] = plugins.manager.get_active_plugins().values()
    assert isinstance(active, SamplePlugin)


def test_custom_hook_registration():

    hook_name = 'some_hook'
    with pytest.raises(LookupError):
        gossip.get_hook(hook_name)

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass
    p = MyPlugin()
    plugins.manager.install(p, activate=True)
    registrations = gossip.get_hook(hook_name).get_registrations()
    assert 1 == len(registrations)
    [r] = registrations
    if PY2:
        assert r.func.__func__ is MyPlugin.unknown.__func__
    else:
        assert r.func.__func__ is MyPlugin.unknown

    # make sure we deactivate properly as well
    plugins.manager.deactivate(p)
    assert not gossip.get_hook(hook_name).get_registrations()


def test_register_invalid_hook():

    initially_installed = list(plugins.manager.get_installed_plugins())

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert list(plugins.manager.get_installed_plugins()) == initially_installed


def test_register_custom_hooks_strict_group():

    initially_installed = list(plugins.manager.get_installed_plugins())

    hook_name = "some_group.some_hook"
    gossip.get_or_create_group("some_group").set_strict()

    class MyPlugin(PluginInterface):

        def get_name(self):
            return "plugin"

        @plugins.registers_on(hook_name)
        def unknown(self):
            pass

    with pytest.raises(IncompatiblePlugin):
        plugins.manager.install(MyPlugin(), activate=True)

    assert list(plugins.manager.get_installed_plugins()) == initially_installed


def test_builtin_plugins_hooks_start_condition():
    "make sure that all hooks are either empty, or contain callbacks marked with `slash.<identifier>`"
    for hook_name, hook in hooks.get_all_hooks():
        for registration in hook.get_registrations():
            assert registration.token.startswith('slash.'), 'Callback {0}.{1} is not a builtin!'.format(hook_name, identifier)

def test_builtin_plugins_are_installed():
    installed = plugins.manager.get_installed_plugins()
    assert installed
    for filename in os.listdir(os.path.join(os.path.dirname(plugins.__file__), "builtin")):
        if filename.startswith("_") or filename.startswith(".") or not filename.endswith(".py"):
            continue
        assert filename[:(-3)] in installed


def test_cannot_install_incompatible_subclasses(no_plugins):

    class Incompatible(object):
        pass
    for invalid in (Incompatible, Incompatible(), PluginInterface, object(), 1, "string"):
        with pytest.raises(IncompatiblePlugin):
            plugins.manager.install(invalid)
    assert plugins.manager.get_installed_plugins() == {}


def test_install_uninstall(no_plugins):
    plugin_name = "some_plugin_name"

    class CustomPlugin(PluginInterface):

        def get_name(self):
            return plugin_name
    with pytest.raises(LookupError):
        plugins.manager.get_plugin(plugin_name)
    plugin = CustomPlugin()
    plugins.manager.install(plugin)
    assert plugins.manager.get_plugin(plugin_name) is plugin
    plugins.manager.uninstall(plugin)
    with pytest.raises(LookupError):
        plugins.manager.get_plugin(plugin_name)






import pytest

@pytest.mark.parametrize('with_without', ['with'])
def test_with_without_nonexisting_plugin(suite, with_without):

    summary = suite.run(additional_args=['--{0}-nonexisting-plugin'.format(with_without)], verify=False)
    assert summary.exit_code != 0



import pytest
import slash
from slash.plugins import PluginInterface


@pytest.fixture(autouse=True)
def restore_plugins(request):
    request.addfinalizer(slash.plugins.manager.install_builtin_plugins)
    request.addfinalizer(slash.plugins.manager.uninstall_all)


def test_plugin_config():

    assert 'sample' not in slash.config['plugin_config']
    value = object()

    class Plugin(PluginInterface):

        def get_name(self):
            return 'sample'

        def get_config(self):
            return {
                'values': {
                    'value_1': value,
                }}

    slash.plugins.manager.install(Plugin())

    assert slash.config.root.plugin_config.sample.values.value_1 is value

    slash.plugins.manager.uninstall('sample')

    assert 'sample' not in slash.config['plugin_config']



import sys

import pytest

from slash.exceptions import INTERRUPTION_EXCEPTIONS
from slash.utils import debug


def test_debug_if_needed_regular_exception(replaced_checkpoint, exc_info, debug_enabled):
    debug.debug_if_needed(exc_info)
    assert replaced_checkpoint.called

def test_debug_if_needed_not_called(replaced_checkpoint, skipped_exc_info, debug_enabled):
    debug.debug_if_needed(skipped_exc_info)
    assert not replaced_checkpoint.called


@pytest.fixture
def debug_enabled(config_override):
    config_override('debug.enabled', True)


@pytest.fixture(params=(SystemExit,) + INTERRUPTION_EXCEPTIONS)
def skipped_exc_info(request):
    try:
        raise request.param()
    except:
        return sys.exc_info()

@pytest.fixture
def exc_info():
    try:
        1/0
    except:
        return sys.exc_info()

@pytest.fixture
def replaced_checkpoint(checkpoint, forge):
    forge.replace_with(debug, 'launch_debugger', checkpoint)
    return checkpoint



from pyparsing import infixNotation, opAssoc, Word, alphanums


class Include(object):

    def __init__(self, t):
        super(Include, self).__init__()
        self.only_tags = False
        self.pattern = t[0]
        if self.pattern.startswith('tag:'):
            self.only_tags = True
            self.pattern = self.pattern.split(':', 1)[1]

    def matches(self, metadata):
        if metadata.tags.matches_pattern(self.pattern):
            return True
        if self.only_tags:
            return False
        return self.pattern in metadata.address

    def __repr__(self):
        return '<{0}{1}>'.format(self.pattern, ' (only tags)' if self.only_tags else '')


class BinaryMatching(object):

    aggregator = None

    def __init__(self, t):
        super(BinaryMatching, self).__init__()
        self.matchers = t[0][0::2]

    def matches(self, metadata):
        return self.aggregator(matcher.matches(metadata) for matcher in self.matchers)  # pylint: disable=not-callable


class AndMatching(BinaryMatching):
    aggregator = all


class OrMatching(BinaryMatching):
    aggregator = any


class Exclude(object):

    def __init__(self, t):
        super(Exclude, self).__init__()
        self.matcher = t[0][1]

    def matches(self, metadata):
        return not self.matcher.matches(metadata)


matcher = Word(alphanums + '._,-=:')
matcher.setParseAction(Include)

boolExpr = infixNotation(matcher, [
    ("not", 1, opAssoc.RIGHT, Exclude),
    ("and", 2, opAssoc.LEFT, AndMatching),
    ("or", 2, opAssoc.LEFT, OrMatching),
])


class Matcher(object):

    def __init__(self, pattern):
        super(Matcher, self).__init__()
        self._matcher = boolExpr.parseString(pattern)[0]

    def __repr__(self):
        return repr(self._matcher)

    def matches(self, metadata):
        return self._matcher.matches(metadata)



import logbook

import pytest
from slash._compat import StringIO
from slash.reporting.console_reporter import ConsoleReporter


def test_console_reporter(suite, level, config_override):
    config_override('log.console_level', level)
    summary = suite.run()
    suite.add_test().when_run.raise_exception()
    assert summary.get_console_output()


def test_silence_manual_errors(suite, suite_test, config_override):
    suite_test.append_line('slash.add_error("msg")')
    suite_test.expect_error()
    config_override('log.show_manual_errors_tb', False)
    output = suite.run().get_console_output()
    assert 'slash.add_error' not in output

def test_num_collected_printed_once(suite, suite_test):
    assert len(suite) > 1
    output = suite.run().get_console_output()
    assert output.count('collected') == 1


@pytest.fixture(params=list(range(logbook.DEBUG, logbook.CRITICAL + 1)))
def level(request):
    return request.param

@pytest.fixture
def stream():
    return StringIO()



from contextlib import contextmanager
import logbook

from sentinels import Sentinel

_logger = logbook.Logger(__name__)

from .. import hooks
from ..ctx import context
from ..exception_handling import handling_exceptions


_LAST_SCOPE = Sentinel('LAST_SCOPE')
_DEDUCE = Sentinel('DEDUCE')



class CleanupManager(object):

    def __init__(self):
        super(CleanupManager, self).__init__()
        self._scope_stack = []
        self._scopes_by_name = {}

    def add_cleanup(self, _func, *args, **kwargs):
        """
        Adds a cleanup function to the cleanup stack. Cleanups are executed in a LIFO order.

        Positional arguments and keywords are passed to the cleanup function when called.

        :param critical: If True, this cleanup will take place even when tests are interrupted by the user (Using Ctrl+C for instance)
        :param success_only: If True, execute this cleanup only if no errors are encountered
        :param scope: Scope at the end of which this cleanup will be executed
        :param args: positional arguments to pass to the cleanup function
        :param kwargs: keyword arguments to pass to the cleanup function
        """

        critical = kwargs.pop('critical', False)
        success_only = kwargs.pop('success_only', False)
        scope_name = kwargs.pop('scope', None)
        if scope_name is None:
            scope = self._scope_stack[-1]
        else:
            scope = self._scopes_by_name[scope_name][-1]

        new_kwargs = kwargs.pop('kwargs', {}).copy()
        new_args = list(kwargs.pop('args', ()))
        if args or kwargs:
            _logger.warning(
                'Passing *args/**kwargs to slash.add_cleanup is deprecated', frame_correction=+2)
            new_args.extend(args)
            new_kwargs.update(kwargs)

        added = _Cleanup(_func, new_args, new_kwargs, critical=critical, success_only=success_only)
        scope.cleanups.append(added)
        return _func

    @contextmanager
    def scope(self, scope):
        self.push_scope(scope)
        try:
            yield
        finally:
            self.pop_scope(scope)

    @property
    def latest_scope(self):
        return self._scope_stack[-1]

    def push_scope(self, scope_name):
        _logger.trace('CleanupManager: pushing scope {0!r}', scope_name)
        scope = _Scope(scope_name)
        self._scope_stack.append(scope)
        self._scopes_by_name.setdefault(scope_name, []).append(scope)

    def pop_scope(self, scope_name, in_failure=None, in_interruption=None):
        _logger.trace('CleanupManager: popping scope {0!r} (failure: {1}, interrupt: {2})', scope_name, in_failure, in_interruption)
        scope = self._scope_stack[-1]
        assert scope.name == scope_name, 'Attempted to pop scope {0!r}, but current scope is {1!r}'.format(scope_name, scope.name)
        self._scope_stack.pop()
        self._scopes_by_name[scope_name].pop()

        self.call_cleanups(
            scope=scope,
            in_failure=in_failure, in_interruption=in_interruption)

    def call_cleanups(self, scope=_LAST_SCOPE, in_failure=False, in_interruption=False):

        _logger.trace('Calling cleanups of scope {0.name!r} (failure={1}, interrupt={2})', scope, in_failure, in_interruption)

        if scope is _LAST_SCOPE:
            scope = self._scope_stack[-1]
            _logger.trace('Deducing last scope={0.name!r}', scope)

        if scope.name == 'test': # pylint: disable=no-member
            with handling_exceptions():
                hooks.before_test_cleanups()  # pylint: disable=no-member

        stack = scope.cleanups # pylint: disable=no-member
        while stack:
            cleanup = stack.pop()
            if in_interruption and not cleanup.critical:
                continue
            if (in_failure or in_interruption) and cleanup.success_only:
                continue
            with handling_exceptions(swallow=True):
                _logger.trace("Calling cleanup: {0}", cleanup)
                cleanup()



class _Cleanup(object):

    def __init__(self, func, args, kwargs, critical=False, success_only=False):
        assert not (success_only and critical)
        super(_Cleanup, self).__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.critical = critical
        self.success_only = success_only
        self.result = context.result

    def __call__(self):
        try:
            return self.func(*self.args, **self.kwargs)
        except Exception:
            self.result.add_exception()
            raise

    def __repr__(self):
        return "{0} ({1},{2})".format(self.func, self.args, self.kwargs)


class _Scope(object):

    def __init__(self, name):
        super(_Scope, self).__init__()
        self.name = name
        self.cleanups = []



import itertools
from confetti import Metadata

_dest_generator = ("dest_{0}".format(i) for i in itertools.count())

# for use with callbacks
def _set_true(_):
    return True

def _set_false(_):
    return False

def _increase(value):
    return value + 1

def _decrease(value):
    return value - 1

class _Cmdline(object):
    def __init__(self, arg=None, on=None, off=None, increase=None, decrease=None, metavar="PARAM", required=False, append=None, increase_doc=None, decrease_doc=None):
        super(_Cmdline, self).__init__()
        dest = next(_dest_generator)
        self.callback_dest = dest + ":callbacks"
        self.arg_dest = dest + ":arg"
        self.arg = arg
        self.on = on
        self.off = off
        self.required = required
        self.increase = increase
        self.decrease = decrease
        self.increase_doc = increase_doc
        self.decrease_doc = decrease_doc
        self.metavar = metavar
        self.append = append

    def configure_parser(self, parser, path, node):
        """
        Add all required flags to a parser to support updating the config value from commandline
        """
        description = node.metadata.get('doc', path)
        if self.arg is not None:
            parser.add_argument(self.arg,
                                dest=self.arg_dest,
                                metavar=self.metavar,
                                default=None,
                                required=self.required,
                                help=description)
        if self.append is not None:
            parser.add_argument(self.append,
                                dest=self.arg_dest,
                                action="append",
                                metavar=self.metavar,
                                default=None,
                                help=description)

        self._add_arg(parser, self.on, callback=_set_true,
                      description="Turn on " + description)
        self._add_arg(parser, self.off, callback=_set_false,
                      description="Turn off " + description)
        self._add_arg(parser, self.increase, callback=_increase,
                      description=self.increase_doc or ("Increase " + description))
        self._add_arg(parser, self.decrease, callback=_decrease,
                      description=self.decrease_doc or ("Decrease " + description))

    def _add_arg(self, parser, flag, callback, description):
        if flag is None:
            return
        parser.add_argument(flag, dest=self.callback_dest, action="append_const", const=callback, help=description)

    def update_value(self, value, args):
        """
        given a previous configuration value and the parsed args, return the new value
        """
        arg_value = getattr(args, self.arg_dest, None)
        if arg_value is not None:
            if self.append:
                value = list(value) + (arg_value or [])
            else:
                value = arg_value
        callbacks = getattr(args, self.callback_dest, None)
        if callbacks:
            for callback in callbacks:
                value = callback(value)
        return value

def Cmdline(**kwargs):
    return Metadata(cmdline=_Cmdline(**kwargs))

def Doc(msg):
    return Metadata(doc=msg)



import sys

from .markers import repeat
from .metadata import Metadata
from ..conf import config


class RunnableTestFactory(object):

    def __init__(self, param=None):
        super(RunnableTestFactory, self).__init__()
        self._param = param
        self._factory_name = self._filename = self._module_name = None

    def set_factory_name(self, name):
        self._factory_name = name

    def get_factory_name(self):
        returned = self._factory_name
        if returned is None:
            returned = self._param.__name__
        assert returned
        return returned

    def set_module_name(self, module_name):
        self._module_name = module_name

    def get_module_name(self):
        returned = self._module_name
        if returned is None:
            returned = self._param.__module__
        assert returned
        return returned

    def set_filename(self, filename):
        self._filename = filename

    def get_filename(self):
        returned = self._filename
        if returned is None:
            returned = sys.modules[self._param.__module__].__file__
        if returned.endswith('.pyc'):
            returned = returned[:-1]
        assert returned
        return returned

    def generate_tests(self, fixture_store):
        """
        Generates :class:`.RunnableTest` instances to run

        Do not override this method directly. Use :func:`.RunnableTestFactory._generate_tests` instead.
        """
        for address_in_factory, test in self._generate_tests(fixture_store):
            assert test.__slash__ is None
            test.__slash__ = Metadata(self, test, address_in_factory)
            yield test

    def _generate_tests(self, fixture_store):
        raise NotImplementedError()  # pragma: no cover

    def _get_num_repetitions(self, func):
        return repeat.get_value(func, 1) * config.root.run.repeat_each



import functools


def mark(name, value, append=False):
    """Marks an object with a given mark (name/value pair)
    """
    return functools.partial(_mark, name=name, value=value, append=append)

def get_marks(obj):
    returned = _get_marks(obj)
    if returned is not None:
        returned = returned.copy()
    return returned

def try_get_mark(obj, mark_name, default=None):
    """Tries getting a specific mark by name from an object, returning None if no such mark is found
    """
    marks = get_marks(obj)
    if marks is None:
        return default
    return marks.get(mark_name, default)

def _get_marks(obj):
    return getattr(obj, _MARKS_CONTAINER_ATTR, None)

def _set_marks(obj, marks):
    setattr(obj, _MARKS_CONTAINER_ATTR, marks)
    return marks

def _mark(obj, name, value, append):
    marks = _get_marks(obj)
    if marks is None:
        marks = _set_marks(obj, {})
    if append:
        marks.setdefault(name, []).append(value)
    else:
        marks[name] = value
    return obj

_MARKS_CONTAINER_ATTR = "__slashmarks__"



from types import FunctionType
import itertools
import platform
import shutil
import sys
import tempfile

import forge
import logbook
from logbook.compat import LoggingHandler

import gossip
import slash
from slash.conf import config
from slash._compat import ExitStack
from slash.core.runnable_test import RunnableTest
from slash.core.test import TestTestFactory
from slash.core.function_test import FunctionTestFactory

if platform.python_version() < "2.7":
    import unittest2 as unittest
else:
    import unittest


_logger = logbook.Logger(__name__)


class TestCase(unittest.TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self._handler = LoggingHandler()
        self._handler.push_application()
        self.addCleanup(self._handler.pop_application)
        gossip.get_group('slash').set_exception_policy(gossip.RaiseImmediately())
        self.override_config("log.console_level", 10000)  # silence console in tests

    def override_config(self, path, value):
        self.addCleanup(config.assign_path, path, config.get_path(path))
        config.assign_path(path, value)

    def get_new_path(self):
        returned = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, returned)
        return returned

    _forge = None

    @property
    def forge(self):
        if self._forge is None:
            self._forge = forge.Forge()
        return self._forge
    _events = None

    @property
    def events(self):
        if self._events is None:
            self._events = self.forge.create_wildcard_mock()
        return self._events

    def tearDown(self):
        if self._forge is not None:
            self._forge.restore_all_replacements()
            self._forge.verify()
        super(TestCase, self).tearDown()


class NullFile(object):

    def write(self, s):
        pass

    def isatty(self):
        return False

    def flush(self):
        pass


class CustomException(Exception):

    @classmethod
    def do_raise(cls):
        raise cls("Custom exception")


def no_op(*args, **kwargs):
    pass


def run_tests_in_session(test_class_path_or_iterator, session=None):
    with ExitStack() as stack:
        if session is None:
            session = slash.Session()
            stack.enter_context(session)

        test_class_path_or_iterator = make_runnable_tests(test_class_path_or_iterator)

        with session.get_started_context():
            slash.runner.run_tests(test_class_path_or_iterator)
    for result in session.results.iter_test_results():
        for err in itertools.chain(result.get_errors(), result.get_failures(), result.get_skips()):
            _logger.debug("Unsuccessful result: {0}", err)
    return session
run_tests_in_session.__test__ = False


def run_tests_assert_success(test_class_path_or_iterator, session=None):
    session = run_tests_in_session(test_class_path_or_iterator, session=session)
    assert session.results.is_success(), "Run did not succeed"
    return session

run_tests_assert_success.__test__ = False


def make_runnable_tests(thing):
    return slash.loader.Loader().get_runnables(thing)


def without_pyc(filename):
    if filename.endswith('.pyc'):
        return filename[:-1]
    return filename



from __future__ import print_function

import argparse
import sys
from functools import partial

import colorama
import slash
from slash.utils.cli_utils import make_styler

_config_name_style = make_styler(colorama.Fore.WHITE + colorama.Style.BRIGHT) # pylint: disable=no-member
_default_style = make_styler(colorama.Fore.GREEN + colorama.Style.BRIGHT) # pylint: disable=no-member
_INDENT = ' ' * 4

def _parse_args(args):
    parser = argparse.ArgumentParser('slash list-config [options] PATH...')
    return parser.parse_args(args)


def list_config(args, report_stream=sys.stdout):
    _print = partial(print, file=report_stream)

    args = _parse_args(args)

    with slash.Session():
        slash.site.load()

        for name, value in sorted(slash.config.traverse_leaves()):
            _print(_config_name_style(name), '--')
            _print(_INDENT, 'default:', _default_style(value.get_value()))
            if value.metadata and 'doc' in value.metadata:
                _print(_INDENT, value.metadata['doc'])



import inspect
import itertools
from uuid import uuid1

import pytest
import slash
from slash.exceptions import CyclicFixtureDependency, UnresolvedFixtureStore, UnknownFixtures, InvalidFixtureScope
from slash.core.fixtures.parameters import bound_parametrizations_context, get_parametrizations
from slash.core.fixtures.fixture_store import FixtureStore


def test_fixture_initialization_order_is_preserved(store):

    _fixture = lambda f: store.add_fixture(slash.fixture(f))
    generation = itertools.count(1)

    @_fixture
    def fixture1():
        return next(generation)

    @_fixture
    def fixture2():
        return next(generation)

    @_fixture
    def fixture3():
        return next(generation)

    store.resolve()
    d = store.get_fixture_dict(['fixture1', 'fixture2', 'fixture3'])
    for i in [1, 2, 3]:
        assert d['fixture{0}'.format(i)] == i


def test_fixture_id_remains_even_when_context_popped(store):

    @slash.fixture
    def fixture0():
        pass

    store.push_namespace()
    store.add_fixture(fixture0)
    store.resolve()

    fixture_obj = store.get_fixture_by_name('fixture0')
    assert fixture_obj.fixture_func is fixture0
    fixture_id = fixture_obj.info.id
    assert store.get_fixture_by_id(fixture_id) is fixture_obj

    store.pop_namespace()

    with pytest.raises(UnknownFixtures):
        store.get_fixture_by_name('fixture0')

    assert store.get_fixture_by_id(fixture_id) is fixture_obj


def test_variations_no_names(store):
    assert list(store.iter_parametrization_variations([])) == [{}]


def test_adding_fixture_twice_to_store(store):

    @slash.fixture
    def fixture0():
        pass

    store.add_fixture(fixture0)
    fixtureobj = store.get_fixture_by_name('fixture0')
    store.add_fixture(fixture0)
    assert store.get_fixture_by_name('fixture0') is fixtureobj


def test_fixture_store_namespace_repr(store):

    @store.add_fixture
    @slash.fixture
    def fixture0():
        pass

    ns = store.get_current_namespace()
    assert str(ns) == repr(ns)
    assert repr(ns) == 'Fixtures: fixture0'


def test_fixture_parameters(store):

    @store.add_fixture
    @slash.fixture
    def value(x, a):
        assert a == 'a', 'Fixture got unexpectedly overriden by parameter'
        return x

    @store.add_fixture
    @slash.fixture
    def a():
        return 'a'

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('a', [1, 2, 3])
    def x(a, b):
        return (a, b)

    @store.add_fixture
    @slash.parametrize('b', [4, 5, 6])
    @slash.fixture
    def b(b):
        return b

    store.resolve()

    variations = list(_get_all_values(store, 'value'))
    assert set(variations) == set(itertools.product([1, 2, 3], [4, 5, 6]))


def test_variation_equality(store):

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('a', [1, 2, 3])
    def fixture(a):
        pass

    store.resolve()

    prev_variation = None
    for variation in store.iter_parametrization_variations(fixture_ids=[store.get_fixture_by_name('fixture').info.id]):
        assert variation == variation
        assert not (variation != variation)
        assert variation != prev_variation
        assert not (variation == prev_variation)
        prev_variation = variation


def _get_all_values(store, fixture_name):
    returned = []
    for variation in store.iter_parametrization_variations(fixture_ids=[store.get_fixture_by_name(fixture_name).info.id]):
        store.push_scope('test')
        with bound_parametrizations_context(variation):
            returned.append(
                store.get_fixture_dict([fixture_name])[fixture_name])
        store.pop_scope('test', in_failure=False, in_interruption=False)
    return returned


@pytest.mark.parametrize('scopes', [('module', 'test'), ('session', 'module'), ('session', 'test')])
def test_wrong_scoping(store, scopes):

    @store.add_fixture
    @slash.fixture(scope=scopes[0])
    def fixture1(fixture2):
        pass

    @store.add_fixture
    @slash.fixture(scope=scopes[1])
    def fixture2():
        pass

    with pytest.raises(InvalidFixtureScope):
        store.resolve()


def test_this_argument(store):

    @store.add_fixture
    @slash.fixture
    def sample(this, other):
        assert this.name == 'sample'
        assert other == 'ok_other'
        return 'ok_sample'

    @store.add_fixture
    @slash.fixture
    def other(this):
        assert this.name == 'other'
        return 'ok_other'

    store.resolve()

    assert store.get_fixture_dict(['sample']) == {
        'sample': 'ok_sample',
    }


def test_fixture_store_unresolved(store):

    @store.add_fixture
    @slash.fixture
    def some_fixture(a, b, c):
        return a + b + c

    with pytest.raises(UnresolvedFixtureStore):
        store.get_fixture_dict(['some_fixture'])


def test_fixture_store_resolve_missing_fixtures(store):

    @store.add_fixture
    @slash.fixture
    def some_fixture(a, b, c):
        return a + b + c

    with pytest.raises(UnknownFixtures):
        store.resolve()


def test_get_all_needed_parametrization_ids(store):

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('param', [1, 2, 3])
    def fixture1(param):
        pass

    @store.add_fixture
    @slash.fixture
    def fixture2(fixture1):
        pass

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('param', [4, 5, 6])
    def fixture3(fixture2, param):
        pass

    fixtureobj = store.get_fixture_by_id(fixture3.__slash_fixture__.id)

    with pytest.raises(UnresolvedFixtureStore):
        store.get_all_needed_parametrization_ids(fixtureobj)

    store.resolve()

    assert len(set(store.get_all_needed_parametrization_ids(fixtureobj))) == 2


def test_get_all_needed_parametrization_ids_of_parametrization(store):

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('param', [1, 2, 3])
    def fixture1(param):
        pass

    fixtureobj = store.get_fixture_by_id(fixture1.__slash_fixture__.id)
    [param_fixtureobj] = get_parametrizations(fixture1)

    with pytest.raises(UnresolvedFixtureStore):
        store.get_all_needed_parametrization_ids(fixtureobj)

    store.resolve()

    needed = set(store.get_all_needed_parametrization_ids(param_fixtureobj))
    assert len(needed) == 1
    assert needed == set([param_fixtureobj.info.id])



def test_fixture_store_iter_parametrization_variations_missing_fixtures(store):

    def test_func(needed_fixture):
        pass

    with pytest.raises(UnknownFixtures):
        list(store.iter_parametrization_variations(funcs=[test_func]))


def test_fixture_store_iter_parametrization_variations_unresolved(store):

    @store.add_fixture
    @slash.fixture
    @slash.parametrize('x', [1, 2, 3])
    def needed_fixture(x):
        pass

    def test_func(needed_fixture):
        pass

    with pytest.raises(UnresolvedFixtureStore):
        list(store.iter_parametrization_variations(funcs=[test_func]))



def test_fixture_dependency(store):
    counter = itertools.count()

    @store.add_fixture
    @slash.fixture
    def fixture1(fixture2):
        assert fixture2 == 'fixture2_value_0'
        return 'fixture1_value_{0}'.format(next(counter))

    @store.add_fixture
    @slash.fixture
    def fixture2():
        return 'fixture2_value_{0}'.format(next(counter))

    store.resolve()

    assert store.get_fixture_dict(['fixture1', 'fixture2']) == {
        'fixture1': 'fixture1_value_1',
        'fixture2': 'fixture2_value_0',
    }


def test_nested_store_resolution_activation(store):

    store.push_namespace()

    @store.add_fixture
    @slash.fixture
    def fixture0():
        return '0'

    store.push_namespace()

    @store.add_fixture
    @slash.fixture
    def fixture1(fixture0):
        assert fixture0 == '0'
        return '1'

    store.push_namespace()

    @store.add_fixture
    @slash.fixture
    def fixture2(fixture1, fixture0):
        assert fixture0 == '0'
        assert fixture1 == '1'
        return '2'

    store.resolve()

    assert store.get_fixture_dict(['fixture2']) == {
        'fixture2': '2'
    }

    store.pop_namespace()

    with pytest.raises(UnknownFixtures):
        store.get_fixture_dict(['fixture2'])


def test_fixture_dependency_cycle():
    store = FixtureStore()

    @store.add_fixture
    @slash.fixture
    def fixture1(fixture2):
        return 1

    @store.add_fixture
    @slash.fixture
    def fixture2(fixture3):
        return 2

    @store.add_fixture
    @slash.fixture
    def fixture3(fixture1):
        return 3

    store.resolve()

    with pytest.raises(CyclicFixtureDependency):
        store.get_fixture_dict(['fixture1'])


def test_fixture_decorator():

    def func(a, b, c):
        pass

    assert not hasattr(func, '__slash_fixture__')

    assert slash.fixture(func) is func

    assert func.__slash_fixture__ is not None


def test_fixture_decorator_multiple_calls(fixture_func):
    fixture_info = fixture_func.__slash_fixture__
    assert slash.fixture(slash.fixture(fixture_func)) is fixture_func

    assert fixture_func.__slash_fixture__ is fixture_info


def test_fixture_required_fixtures(fixture_func):
    assert fixture_func.__slash_fixture__.required_args == inspect.getargspec(
        fixture_func).args


def test_fixture_name(fixture_func, fixture_func_name):
    assert fixture_func.__slash_fixture__.name == fixture_func_name


def test_fixture_store_add(fixture_func, fixture_func_name):
    f = FixtureStore()
    assert f.add_fixture(fixture_func) is fixture_func
    assert f.get_fixture_by_name(
        fixture_func_name).fixture_func is fixture_func

def test_nofixtures_decorator(store):

    @slash.nofixtures
    def func1(a, b, c):
        pass

    [var] = store.iter_parametrization_variations(funcs=[func1])
    assert var == {}

def test_nofixtures_decorator_methods(store):

    class TestClass(slash.Test):
        @slash.nofixtures
        def before(self, a, b, c):
            pass

    [var] = store.iter_parametrization_variations(methods=[('before', TestClass.before)])
    assert var == {}



@pytest.fixture(params=[True, False])
def fixture_func(request, fixture_func_name):

    def func(a, b, c, d):
        pass

    use_function_name = request.param
    if use_function_name:
        func.__name__ = fixture_func_name
        return slash.fixture(func)
    return slash.fixture(name=fixture_func_name)(func)


@pytest.fixture
def fixture_func_name():
    return str(uuid1()).replace('-', '_')


@pytest.fixture
def store():
    return FixtureStore()

@pytest.fixture(autouse=True)
def non_null_ctx(request):
    slash.ctx.push_context()
    request.addfinalizer(slash.ctx.pop_context)



@pytest.fixture
def cleanup_map():
    return {}


@pytest.fixture
def test_scoped_fixture(store, cleanup_map):

    return _get_scoped_fixture('test', store, cleanup_map)


@pytest.fixture
def module_scoped_fixture(store, cleanup_map):

    return _get_scoped_fixture('module', store, cleanup_map)


@pytest.fixture
def session_scoped_fixture(store, cleanup_map):

    return _get_scoped_fixture('session', store, cleanup_map)


def _get_scoped_fixture(scope, store, cleanup_map):
    @store.add_fixture
    @slash.fixture(scope=scope, name='{0}_scoped_fixture'.format(scope))
    def fixture(this):

        @this.add_cleanup
        def cleanup():
            cleanup_map[this.name] = True

        return ok(this.name)
    return fixture


def ok(s):
    return 'ok_{0}'.format(s)



import pytest



import slash

from .utils.suite_writer import Suite

_MODULE_SCOPE_ADDER = 'slash.add_cleanup({0}, scope="module")'


def test_cleanups_from_test_start(suite):

    num_tests = 5
    for i in range(num_tests):
        suite.add_test()

    events = []
    test_ids = []

    @slash.hooks.register
    def test_start():
        test_id = slash.context.test.__slash__.id
        test_ids.append(test_id)
        events.append(('test_start', test_id))
        @slash.add_cleanup
        def cleanup():
            events.append(('test_cleanup', test_id))

    suite.run()

    assert len(events) == len(suite) * 2 == num_tests * 2
    expected = []
    for test_id in test_ids:
        expected.append(('test_start', test_id))
        expected.append(('test_cleanup', test_id))
    assert events == expected


def test_module_scope(scoped_suite, file1_tests, file2_tests):
    file1_test = file1_tests[0]
    file1_end = file1_tests[-1].add_deferred_event(decorator='slash.add_cleanup')
    file1_test_cleanup = file1_test.add_deferred_event(adder=_MODULE_SCOPE_ADDER)

    summary = scoped_suite.run()
    assert summary.events[file1_end].timestamp < summary.events[file1_test_cleanup].timestamp


def test_test_scoped_cleanups_in_session(checkpoint):
    # with scoped cleanups, and the default being 'test', there is a special meaning
    # for cleanups registered outside of tests....
    with slash.Session() as s:
        slash.add_cleanup(checkpoint)
        assert not checkpoint.called
        with s.get_started_context():
            pass

        assert not checkpoint.called
    assert checkpoint.called


def test_errors_associated_with_correct_result(scoped_suite, file1_tests, file2_tests):
    file1_test = file1_tests[0]
    file1_test_cleanup = file1_test.add_deferred_event(adder=_MODULE_SCOPE_ADDER, extra_code=['assert 1 == 2'])
    file1_test.expect_failure()

    scoped_suite.run()


@pytest.fixture
def scoped_suite(suite, file1_tests, file2_tests):
    return suite

@pytest.fixture
def suite():
    return Suite()


@pytest.fixture
def file1_tests(suite):
    file1 = suite.add_file()
    return [file1.add_function_test() for i in range(5)]

@pytest.fixture
def file2_tests(suite):
    file2 = suite.add_file()
    return [file2.add_function_test() for i in range(3)]



import functools

import logbook

from ..ctx import context
from ..utils.python import call_all_raise_first
from ..exceptions import SkipTest, INTERRUPTION_EXCEPTIONS

_logger = logbook.Logger(__name__)

class ScopeManager(object):

    def __init__(self, session):
        super(ScopeManager, self).__init__()
        self._session = session
        self._scopes = []
        self._last_module = self._last_test = None

    def begin_test(self, test):
        test_module = test.__slash__.module_name
        assert test_module

        if self._last_module is None:
            self._push_scope('session')


        if self._last_module != test_module:
            if self._last_module is not None:
                _logger.trace('Module scope has changed. Popping previous module scope')
                self._pop_scope('module', in_failure=False, in_interruption=False)
            assert self._scopes[-1] != 'module'
            self._push_scope('module')
        self._last_module = test_module
        self._push_scope('test')
        self._last_test = test

    def end_test(self, test, next_test, exc_info):
        assert test == self._last_test

        exc_type = exc_info[0]
        in_failure = exc_type is not None and not issubclass(exc_type, SkipTest)
        if context.result is not None:
            in_failure = in_failure or context.result.is_error() or context.result.is_failure()
        kw = {'in_failure': in_failure, 'in_interruption': exc_type in INTERRUPTION_EXCEPTIONS}

        self._pop_scope('test', **kw)

        if next_test is None:
            _logger.trace('No next test. Popping scopes')
            self._pop_scope('module', **kw)
            self._pop_scope('session', **kw)

    def get_current_stack(self):
        return self._scopes[:]

    def _push_scope(self, scope):
        self._scopes.append(scope)
        _logger.trace('Pushed scope {0}', scope)
        self._session.fixture_store.push_scope(scope)
        self._session.cleanups.push_scope(scope)

    def _pop_scope(self, scope, **kw):
        popped = self._scopes.pop()
        _logger.trace('Popped scope {0} (expected {1})', popped, scope)
        assert popped == scope
        call_all_raise_first([self._session.cleanups.pop_scope, self._session.fixture_store.pop_scope],
                             scope, **kw)

    def flush_remaining_scopes(self, **kw):
        call_all_raise_first([functools.partial(self._pop_scope, s)
                              for s in self._scopes[::-1]], **kw)



from ..interface import PluginInterface
from ...conf import config
from ...ctx import session
from functools import partial
import time
import socket
import requests
import gossip


def _post_notification(url, api_key, title, message):
    response = requests.post(url, {"apikey": api_key, "application": "slash", "event": title, "description": message})
    response.raise_for_status()


def _pushbullet_notification(api_key, title, message):
    data = {
        "type": "note",
        "title": title,
        "body": message
    }
    response = requests.post(
        "https://api.pushbullet.com/api/pushes",
        data=data,
        auth=(api_key, ""))
    response.raise_for_status()


class Message(object):
    def __init__(self, title, body):
        self.title = title
        self.body = body


class Plugin(PluginInterface):
    def get_name(self):
        return "notifications"

    def get_config(self):
        return {
            "prowl_api_key" : None,
            "nma_api_key" : None,
            "pushbullet_api_key": None,
            "notification_threshold": 5,
        }

    def session_start(self):
        self._session_start_time = time.time() #pylint: disable=W0201

    def session_end(self):
        if time.time() - self._session_start_time < config.root.plugin_config.notifications.notification_threshold:
            return

        result = "successfully" if session.results.is_success() else "unsuccessfully"
        hostname = socket.gethostname().split(".")[0]
        body = "{0}\n\n{1}".format(session.results, session.id)

        message = Message("Slash session in {0} ended {1}".format(hostname, result), body)
        gossip.trigger("slash-gossip.prepare_notification", message=message)

        self._notify(message.title, message.body)

    def _notify(self, title, message):
        for func, api_key in [
                (partial(_post_notification, "https://prowl.weks.net/publicapi/add"), config.root.plugin_config.notifications.prowl_api_key),
                (partial(_post_notification, "https://www.notifymyandroid.com/publicapi/notify"), config.root.plugin_config.notifications.nma_api_key),
                (_pushbullet_notification, config.root.plugin_config.notifications.pushbullet_api_key),
                ]:
            if api_key is None:
                continue

            func(api_key, title, message)



from uuid import uuid1

import pytest
import slash
from slash import context, Session
from slash.ctx import Context, ContextAttributeProxy


def test_null_context():
    assert context.test_id is None
    assert context.result is None

def test_session_context_result():
    with Session() as s:
        assert context.session is s
        assert context.result is s.results.global_result

def test_test_context_result(suite):
    for test in suite:
        test.append_line('assert slash.context.result is slash.session.results[slash.context.test]')

    @slash.hooks.result_summary.register
    def assert_result_back_to_normal():
        assert context.result is context.session.results.global_result

    assert suite.run().ok()

def test_cannot_pop_bottom():
    assert len(context._stack) == 1
    with pytest.raises(RuntimeError):
        context.pop()

def test_push_context(loaded_context):
    test_id = loaded_context.test_id = "some_test_id"
    assert context.test_id == test_id

def test_object_proxy_getattr(contextobj, realobj):
    assert contextobj.attr is realobj.attr
    assert contextobj.__attr__ is realobj.__attr__
    assert not hasattr(contextobj, "nonexisting")
    assert not hasattr(contextobj, "__nonexisting__")

def test_object_proxy_eq(contextobj, realobj):
    assert contextobj == realobj

def test_object_proxy_ne(contextobj, realobj):
    assert not (contextobj != realobj)

def test_object_proxy_str_repr(contextobj, realobj):
    assert str(contextobj) == str(realobj)
    assert repr(contextobj) == repr(realobj)

@pytest.fixture
def contextobj(loaded_context, realobj):
    setattr(loaded_context, "obj", realobj)
    return ContextAttributeProxy("obj")

@pytest.fixture
def realobj():
    class Object(object):

        __attr__ = object()
        attr = object()

    return Object()

@pytest.fixture
def loaded_context():
    returned = Context()
    context.push(returned)
    return returned

@pytest.fixture(autouse=True, scope="function")
def pop_all(request):
    @request.addfinalizer
    def cleanup():
        while len(context._stack) > 1:
            context.pop()



import pytest
import slash


def test_frame_order(error, use_exception):
    code_line = error.traceback.frames[-1].code_line.strip()
    if use_exception:
        assert code_line == '1 / 0'
    else:
        assert code_line == 'get_error_adder()("error message")'


@pytest.fixture
def error(get_error_adder, use_exception):

    def f1():
        try:
            g1()
        except ZeroDivisionError:
            assert use_exception
            get_error_adder()()

    def g1():
        h1()

    def h1():
        if use_exception:
            1 / 0
        else:
            get_error_adder()("error message")

    with slash.Session() as s:
        f1()
    [err] = s.results.global_result.get_errors()
    return err


@pytest.fixture(params=[lambda: slash.add_error, lambda: slash.context.result.add_error])
def get_error_adder(request):
    return request.param


@pytest.fixture(params=[True, False])
def use_exception(request):
    return request.param



import itertools

import pytest
import slash
from sentinels import NOTHING
from slash.core.tagging import NO_TAGS, Tags, get_tags
from slash.loader import Loader

from .utils.suite_writer.method_test import MethodTest


def test_setting_getting_tags(taggable):

    slash.tag('tagname')(taggable)

    assert 'tagname' in get_tags(taggable)
    assert 'other_tags' not in get_tags(taggable)


def test_tags_addition_no_tags_no_tags():
    assert NO_TAGS + NO_TAGS is NO_TAGS


def test_tags_addition_no_tags_regular():
    tags = Tags({'a': 'b'})
    assert NO_TAGS + tags is tags
    assert tags + NO_TAGS is tags


def test_tags_addition_regular():
    tags1 = Tags({'a': 'b'})
    tags2 = Tags({'c': 'd'})
    tags3 = tags1 + tags2
    assert tags3 is not tags1
    assert tags3 is not tags2

    assert tags1._tags == {'a': 'b'}
    assert tags2._tags == {'c': 'd'}
    assert tags3._tags == {'a': 'b', 'c': 'd'}


def test_setting_getting_tags_on_metadata(taggable):

    slash.tag('tagname')(taggable)

    with slash.Session() as s:
        tests = Loader().get_runnables(taggable)
    assert tests
    for t in tests:
        assert 'tagname' in t.__slash__.tags


def test_metadata_tags(suite, suite_test, tagging_strategy, tags):
    tagging_strategy(suite_test, tags)
    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)
    for tag_name, tag_value in tags:
        assert result.test_metadata.tags.has_tag(tag_name)
        assert result.test_metadata.tags[tag_name] == tag_value

# more tags in test_pattern_matching.py

_tagging_strategies = []


def _tagging_strategy(func):
    _tagging_strategies.append(func)
    return func


@_tagging_strategy
def _simple_tagging_strategy(taggable, tags):
    for tag_name, tag_value in tags:
        taggable.add_decorator(_get_slash_tag_string(tag_name, tag_value))


@_tagging_strategy
def _tag_class_only(taggable, tags):
    if isinstance(taggable, MethodTest):
        _simple_tagging_strategy(taggable.cls, tags)
    else:
        _simple_tagging_strategy(taggable, tags)


@_tagging_strategy
def _tag_class_and_method(taggable, tags):
    if not isinstance(taggable, MethodTest):
        return _simple_tagging_strategy(taggable, tags)

    for (tag_name, tag_value), taggable in zip(tags, itertools.cycle([taggable.cls, taggable])):
        taggable.add_decorator(_get_slash_tag_string(tag_name, tag_value))


def _get_slash_tag_string(tag_name, tag_value):
    returned = 'slash.tag({0!r}'.format(tag_name)
    if tag_value is not NOTHING:
        returned += ', {0!r}'.format(tag_value)
    returned += ')'
    return returned


@pytest.fixture(params=_tagging_strategies)
def tagging_strategy(request):
    return request.param


@pytest.fixture(params=[
    {'simple_tag_without_value': NOTHING},
    {'single_tag': 'string_value'},
    {'multiple_tags_1': 1.0, 'multiple_tags_2': True, 'multiple_tags_3': ['list', 'of', 'things']},
])
def tags(request):
    return list(request.param.items())


@pytest.fixture(params=['class', 'function'])
def taggable(request):
    if request.param == 'class':
        class TaggableTest(slash.Test):

            def test_1():
                pass
        return TaggableTest
    elif request.param == 'function':
        def test_1():
            pass
        return test_1

    raise NotImplementedError()  # pragma: no cover



import os

import pytest
import slash
from lxml import etree


def test_xunit_plugin(results, xunit_filename):
    assert os.path.exists(xunit_filename), 'xunit file not created'

    schema_root = etree.XML(_XUNIT_XSD)
    schema = etree.XMLSchema(schema_root)

    parser = etree.XMLParser(schema=schema)
    with open(xunit_filename) as f:
        etree.parse(f, parser)


@pytest.fixture
def results(suite, xunit_filename):
    suite.run()

@pytest.fixture
def xunit_filename(tmpdir, request, config_override):
    xunit_filename = str(tmpdir.join('xunit.xml'))
    slash.plugins.manager.activate('xunit')

    slash.config.root.plugins.xunit.filename = xunit_filename

    @request.addfinalizer
    def deactivate():
        slash.plugins.manager.deactivate('xunit')
        assert 'xunit' not in slash.config['plugins']

    return xunit_filename


# Taken from https://gist.github.com/jzelenkov/959290
_XUNIT_XSD = """<?xml version="1.0"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
   elementFormDefault="qualified"
   attributeFormDefault="unqualified">
  <xs:annotation>
    <xs:documentation xml:lang="en">Jenkins xUnit test result schema.
    </xs:documentation>
  </xs:annotation>
  <xs:element name="testsuite" type="testsuite"/>
  <xs:simpleType name="ISO8601_DATETIME_PATTERN">
    <xs:restriction base="xs:dateTime">
      <xs:pattern value="[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:element name="testsuites">
    <xs:annotation>
      <xs:documentation xml:lang="en">Contains an aggregation of testsuite results</xs:documentation>
    </xs:annotation>
    <xs:complexType>
      <xs:sequence>
        <xs:element name="testsuite" minOccurs="0" maxOccurs="unbounded">
          <xs:complexType>
            <xs:complexContent>
              <xs:extension base="testsuite">
                <xs:attribute name="package" type="xs:token" use="required">
                  <xs:annotation>
                    <xs:documentation xml:lang="en">Derived from testsuite/@name in the non-aggregated documents</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="id" type="xs:int" use="required">
                  <xs:annotation>
                    <xs:documentation xml:lang="en">Starts at '0' for the first testsuite and is incremented by 1 for each following testsuite</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
              </xs:extension>
            </xs:complexContent>
          </xs:complexType>
        </xs:element>
      </xs:sequence>
    </xs:complexType>
  </xs:element>
  <xs:complexType name="testsuite">
    <xs:annotation>
      <xs:documentation xml:lang="en">Contains the results of exexuting a testsuite</xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="testcase" minOccurs="0" maxOccurs="unbounded">
        <xs:complexType>
          <xs:choice minOccurs="0">
            <xs:element name="error">
      <xs:annotation>
        <xs:documentation xml:lang="en">Indicates that the test errored.  An errored test is one that had an unanticipated problem. e.g., an unchecked throwable; or a problem with the implementation of the test. Contains as a text node relevant data for the error, e.g., a stack trace</xs:documentation>
      </xs:annotation>
              <xs:complexType>
                <xs:simpleContent>
                  <xs:extension base="pre-string">
                    <xs:attribute name="message" type="xs:string">
                      <xs:annotation>
                        <xs:documentation xml:lang="en">The error message. e.g., if a java exception is thrown, the return value of getMessage()</xs:documentation>
                      </xs:annotation>
                    </xs:attribute>
                    <xs:attribute name="type" type="xs:string" use="required">
                      <xs:annotation>
                        <xs:documentation xml:lang="en">The type of error that occured. e.g., if a java execption is thrown the full class name of the exception.</xs:documentation>
                      </xs:annotation>
                    </xs:attribute>
                  </xs:extension>
                </xs:simpleContent>
              </xs:complexType>
            </xs:element>
            <xs:element name="failure">
      <xs:annotation>
        <xs:documentation xml:lang="en">Indicates that the test failed. A failure is a test which the code has explicitly failed by using the mechanisms for that purpose. e.g., via an assertEquals. Contains as a text node relevant data for the failure, e.g., a stack trace</xs:documentation>
      </xs:annotation>
              <xs:complexType>
                <xs:simpleContent>
                  <xs:extension base="pre-string">
                    <xs:attribute name="message" type="xs:string">
                      <xs:annotation>
                        <xs:documentation xml:lang="en">The message specified in the assert</xs:documentation>
                      </xs:annotation>
                    </xs:attribute>
                    <xs:attribute name="type" type="xs:string" use="required">
                      <xs:annotation>
                        <xs:documentation xml:lang="en">The type of the assert.</xs:documentation>
                      </xs:annotation>
                    </xs:attribute>
                  </xs:extension>
                </xs:simpleContent>
              </xs:complexType>
            </xs:element>
            <xs:element name="skipped">
      <xs:annotation>
        <xs:documentation xml:lang="en">Indicates that the test was skipped. A skipped test is a test which was ignored using framework mechanisms. e.g., @Ignore annotation.</xs:documentation>
      </xs:annotation>
              <xs:complexType>
                <xs:simpleContent>
                  <xs:extension base="pre-string">
                    <xs:attribute name="type" type="xs:string" use="required">
                      <xs:annotation>
                        <xs:documentation xml:lang="en">Skip type.</xs:documentation>
                      </xs:annotation>
                    </xs:attribute>
                  </xs:extension>
                </xs:simpleContent>
              </xs:complexType>
            </xs:element>
          </xs:choice>
          <xs:attribute name="name" type="xs:token" use="required">
            <xs:annotation>
              <xs:documentation xml:lang="en">Name of the test method</xs:documentation>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="classname" type="xs:token" use="required">
            <xs:annotation>
              <xs:documentation xml:lang="en">Full class name for the class the test method is in.</xs:documentation>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="time" type="xs:decimal" use="required">
            <xs:annotation>
              <xs:documentation xml:lang="en">Time taken (in seconds) to execute the test</xs:documentation>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="name" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">Full class name of the test for non-aggregated testsuite documents. Class name without the package for aggregated testsuites documents</xs:documentation>
      </xs:annotation>
      <xs:simpleType>
        <xs:restriction base="xs:token">
          <xs:minLength value="1"/>
        </xs:restriction>
      </xs:simpleType>
    </xs:attribute>
    <xs:attribute name="timestamp" type="ISO8601_DATETIME_PATTERN" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">when the test was executed. Timezone may not be specified.</xs:documentation>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="hostname" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">Host on which the tests were executed. 'localhost' should be used if the hostname cannot be determined.</xs:documentation>
      </xs:annotation>
      <xs:simpleType>
        <xs:restriction base="xs:token">
          <xs:minLength value="1"/>
        </xs:restriction>
      </xs:simpleType>
    </xs:attribute>
    <xs:attribute name="tests" type="xs:int" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">The total number of tests in the suite</xs:documentation>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="failures" type="xs:int" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">The total number of tests in the suite that failed. A failure is a test which the code has explicitly failed by using the mechanisms for that purpose. e.g., via an assertEquals</xs:documentation>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="errors" type="xs:int" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">The total number of tests in the suite that errored. An errored test is one that had an unanticipated problem. e.g., an unchecked throwable; or a problem with the implementation of the test.</xs:documentation>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="skipped" type="xs:int">
      <xs:annotation>
        <xs:documentation xml:lang="en">The total number of tests in the suite that skipped. A skipped test is a test which was ignored using framework mechanisms. e.g., @Ignore annotation.</xs:documentation>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="time" type="xs:decimal" use="required">
      <xs:annotation>
        <xs:documentation xml:lang="en">Time taken (in seconds) to execute the tests in the suite</xs:documentation>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:simpleType name="pre-string">
    <xs:restriction base="xs:string">
      <xs:whiteSpace value="preserve"/>
    </xs:restriction>
  </xs:simpleType>
</xs:schema>
"""



import functools
import itertools

from .._compat import iteritems, izip, xrange
from ..exception_handling import handling_exceptions
from ..exceptions import SkipTest
from .fixtures.parameters import bound_parametrizations_context
from .runnable_test import RunnableTest
from .runnable_test_factory import RunnableTestFactory
from .requirements import get_requirements
from .tagging import get_tags


class TestTestFactory(RunnableTestFactory):

    def __init__(self, testclass):
        super(TestTestFactory, self).__init__(testclass)
        self.testclass = testclass

    def _generate_tests(self, fixture_store):
        if is_abstract_base_class(self.testclass):
            return

        for test_method_name in dir(self.testclass):
            if not test_method_name.startswith("test"):
                continue

            for fixture_variation in self._iter_parametrization_variations(test_method_name, fixture_store):
                for _ in xrange(self._get_num_repetitions(getattr(self.testclass, test_method_name))):
                    case = self.testclass(
                        test_method_name,
                        fixture_store=fixture_store,
                        fixture_namespace=fixture_store.get_current_namespace(),
                        fixture_variation=fixture_variation,
                    )
                    if self.testclass.__slash_skipped__:
                        case.run = functools.partial(SkipTest.throw, self.testclass.__slash_skipped_reason__)
                    yield case._get_address_in_factory(), case  # pylint: disable=protected-access

    def _iter_parametrization_variations(self, test_method_name, fixture_store):
        return fixture_store.iter_parametrization_variations(methods=itertools.chain(
            izip(itertools.repeat('before'), self._get_all_before_methods()),
            izip(itertools.repeat('after'), self._get_all_after_methods()),
            [getattr(self.testclass, test_method_name)],
        ))

    def _get_all_before_methods(self):
        return self._iter_inherited_methods('before')

    def _get_all_after_methods(self):
        return self._iter_inherited_methods('after')

    def _iter_inherited_methods(self, name):

        for cls in self.testclass.__mro__:
            if hasattr(cls, name):
                yield getattr(cls, name)

    def get_unmet_requirements(self):
        raise NotImplementedError() # pragma: no cover


class Test(RunnableTest):
    """
    This is a base class for implementing unittest-style test classes.
    """
    def __init__(self, test_method_name, fixture_store, fixture_namespace, fixture_variation):
        super(Test, self).__init__()
        self._test_method_name = test_method_name
        self._fixture_store = fixture_store
        self._fixture_namespace = fixture_namespace
        self._fixture_variation = fixture_variation

    def get_tags(self):
        return get_tags(type(self)) + get_tags(getattr(type(self), self._test_method_name))

    __slash_skipped__ = False
    __slash_skipped_reason__ = None
    __slash_needed_contexts__ = None


    @classmethod
    def skip_all(cls, reason=None):
        cls.__slash_skipped__ = True
        cls.__slash_skipped_reason__ = reason

    def get_required_fixture_objects(self):
        method = getattr(self, self._test_method_name)
        return self._fixture_store.get_required_fixture_objects(method, namespace=self._fixture_namespace, is_method=True)

    def _get_address_in_factory(self):
        returned = ''
        if self._test_method_name is not None:
            returned += ".{0}".format(self._test_method_name)
        if self._fixture_variation:
            returned += '({0})'.format(self._fixture_variation.representation)
        return returned

    def _get_call_string(self, kwargs):
        if not kwargs:
            return ""
        return "({0})".format(", ".join("{0}={1!r}".format(k, v) for k, v in iteritems(kwargs)))

    def get_requirements(self):
        return get_requirements(type(self)) + get_requirements(getattr(self, self._test_method_name))

    def run(self):  # pylint: disable=E0202
        """
        .. warning:: Not to be overriden
        """
        method = getattr(self, self._test_method_name)
        with bound_parametrizations_context(self._fixture_variation):
            self._fixture_store.activate_autouse_fixtures_in_namespace(self._fixture_namespace)
            self.before()
            try:
                with handling_exceptions():
                    self._fixture_store.call_with_fixtures(
                        method, namespace=self._fixture_namespace,
                        is_method=True,
                    )
            finally:
                with handling_exceptions():
                    self.after()

    def before(self):
        """
        Gets called before each separate case generated from this test class
        """
        pass
    def after(self):
        """
        Gets called after each separate case from this test class executed, assuming :meth:`.before` was successful.
        """
        pass

    def _format_kwargs(self, kwargs):
        return ", ".join("{0}={1!r}".format(x, y) for x, y in iteritems(kwargs))

def abstract_test_class(cls):
    """
    Marks a class as **abstract**, thus meaning it is not to be run
    directly, but rather via a subclass.
    """
    assert issubclass(cls, Test), "abstract_test_class only operates on slash.Test subclasses"
    cls.__slash_abstract__ = True
    return cls

def is_abstract_base_class(cls):
    """
    Checks if a given class is abstract.

    .. seealso:: :func:`abstract_test_class`
    """
    return bool(cls.__dict__.get("__slash_abstract__", False))



_SLASH_REQUIRES_KEY_NAME = '__slash_requirements__'


def requires(req, message=None):
    """A decorator specifying that the decorated tests requires a certain precondition in order to run

    :param req: Either a function receiving no arguments and returning a boolean, or a boolean specifying whether or not
       the requirement is met
    """
    def decorator(func):
        reqs = getattr(func, _SLASH_REQUIRES_KEY_NAME, None)
        if reqs is None:
            reqs = []
            setattr(func, _SLASH_REQUIRES_KEY_NAME, reqs)
        reqs.append(Requirement(req, message))
        return func
    return decorator


def get_requirements(test):
    return list(getattr(test, _SLASH_REQUIRES_KEY_NAME, []))


class Requirement(object):

    def __init__(self, req, message=None):
        super(Requirement, self).__init__()
        self._req = req
        self._message = message

    def __repr__(self):
        if self._message is not None:
            return self._message
        return repr(self._req)

    def is_met(self):
        if isinstance(self._req, bool):
            return self._req
        return self._req()



import logbook
from confetti import Config
from .utils.conf_utils import Doc, Cmdline

__all__ = ["config"]

config = Config({
    "debug": {
        "debug_skips": False // Doc("Enter pdb also for SkipTest exceptions"),
        "debug_hook_handlers": False // Doc("Enter pdb also for every exception encountered in a hook/callback. Only relevant when debugging is enabled"),
        "enabled": False // Doc("Enter pdb on failures and errors") // Cmdline(on="--pdb"),
    },

    "log": {
        "colorize": False // Doc("Emit log colors to files"),
        "console_theme": {
            'inline-file-end-fail': 'red',
            'inline-file-end-skip': 'yellow',
            'inline-file-end-success': 'green',
            'inline-error': 'red',
            'inline-test-interrupted': 'yellow',
            'error-cause-marker': 'white/bold',
            'fancy-message': 'yellow/bold',
            'frame-local-varname': 'yellow/bold',
            'session-summary-success': 'green/bold',
            'session-summary-failure': 'red/bold',
            'error-separator-dash': 'red',
            'tb-error': 'red/bold',
            'tb-frame-location': 'white/bold',
            'test-additional-details-header': 'black/bold',
            'test-additional-details': 'black/bold',
            'test-error-header': 'white',
            'test-skip-message': 'yellow',
            'tb-line-cause': 'white',
            'tb-line': 'black/bold',
        },

        "console_level": logbook.WARNING // Cmdline(decrease='-v',
                                                    decrease_doc='Make console more verbose (can be specified multiple times)',
                                                    increase='-q',
                                                    increase_doc='Make console less verbose (can be specified multiple times)'),
        "traceback_level": 2 // Doc("Detail level of tracebacks") // Cmdline(arg="--tb"),
        "truncate_console_lines": True // Doc("truncate long log lines on the console") // Cmdline(arg='--truncate-console-lines', metavar='yes/no'),
        "truncate_console_errors": False // Doc("If truncate_console_lines is set, also truncate long log lines, including and above the \"error\" level, on the console"),
        "root": None // Doc("Root directory for logs") // Cmdline(arg="-l", metavar="DIR"),
        "subpath": "{context.session.id}/{context.test_id}/debug.log" // Doc("Path to write logs to under the root"),
        "session_subpath": "{context.session.id}/session.log",
        "last_session_symlink": None // Doc("If set, specifies a symlink path to the last session log file in each run"),
        "last_session_dir_symlink": None // Doc("If set, specifies a symlink path to the last session log directory"),
        "last_test_symlink": None // Doc("If set, specifies a symlink path to the last test log file in each run"),
        "last_failed_symlink": None // Doc("If set, specifies a symlink path to the last failed test log file"),
        "show_manual_errors_tb": True // Doc("Show tracebacks for errors added via slash.add_error"),
        "silence_loggers": [] // Doc("Logger names to silence"),
        "format": None // Doc("Format of the log line, as passed on to logbook. None will use the default format"),
        "console_format": None // Doc("Optional format to be used for console output. Defaults to the regular format"),
        "localtime": False // Doc("Use local time for logging. If False, will use UTC"),
        "unittest_mode": False // Doc("Used during unit testing. Emit all logs to stderr as well as the log files"),
        "unified_session_log": False // Doc("Make the session log file contain all logs, including from tests"),
    },
    "run": {
        "default_sources": [] // Doc("Default tests to run assuming no other sources are given to the runner"),
        "suite_files": [] // Doc("File(s) to be read for lists of tests to be run") // Cmdline(append="-f", metavar="FILENAME"),
        "stop_on_error": False // Doc("Stop execution when a test doesn't succeed") // Cmdline(on="-x"),
        "filter_string": None // Doc("A string filter, selecting specific tests by string matching against their name") // Cmdline(arg='-k', metavar='FILTER'),
        "repeat_each": 1 // Doc("Repeat each test a specified amount of times") // Cmdline(arg='--repeat-each', metavar="NUM_TIMES"),
        "session_state_path": "~/.slash/last_session" // Doc("Where to keep last session serialized data"),
        "project_customization_file_path": "./.slashrc",
        "user_customization_file_path": "~/.slash/slashrc",
    },
    "sentry": {
        "dsn": None // Doc("Possible DSN for a sentry service to log swallowed exceptions. "
                           "See http://getsentry.com for details"),
    },
    "plugins": {
        "search_paths": [] // Doc("List of paths in which to search for plugin modules"),
    },

    "plugin_config": {
        # DO NOT store configuration here. It is intended for dynamically loaded plugins
    },
})



import pytest
import slash

from .utils import run_tests_assert_success, without_pyc


@pytest.mark.parametrize('reprify', [repr, str])
def test_str_repr(warning, reprify):
    assert 'this is a warning' in reprify(warning)

def test_location(warning):
    assert warning.details['filename'] == without_pyc(__file__)

def test_to_dict(warning):
    assert isinstance(warning.to_dict(), dict)

def test_warning_added_hook(suite, suite_test):

    captured = []
    @slash.hooks.register
    def warning_added(warning):
        captured.append(warning)

    suite_test.append_line('slash.logger.warning("message here")')
    suite.run()
    assert captured
    [w] = captured
    assert w.message == 'message here'
    assert isinstance(w.lineno, int)
    assert isinstance(w.filename, str)
    assert w.lineno
    assert w.filename

@pytest.fixture
def warning():
    class SampleTest(slash.Test):
        def test(self):
            slash.logger.warning("this is a warning. Param: {0}", 1)

    session = run_tests_assert_success(SampleTest)

    assert len(session.warnings) == 1

    [warning] = session.warnings
    return warning



import itertools

import pytest
import slash.plugins
from slash.plugins import PluginInterface


@pytest.mark.parametrize('needs_decorate_method', [True, False])
@pytest.mark.parametrize('provides_decorate_method', [True, False])
def test_needs_provides_plugin_name(needs_decorate_method, provides_decorate_method, checkpoint1, checkpoint2):

    @slash.plugins.active
    @_maybe_decorate(slash.plugins.needs('p'), not needs_decorate_method)
    @autoname
    class NeedsPlugin(PluginInterface):

        @_maybe_decorate(slash.plugins.needs('p'), needs_decorate_method)
        def session_start(self):
            checkpoint2()

    @slash.plugins.active
    @_maybe_decorate(slash.plugins.provides('p'), not provides_decorate_method)
    @autoname
    class ProvidesPlugin(PluginInterface):

        @_maybe_decorate(slash.plugins.provides('p'), provides_decorate_method)
        def session_start(self):
            checkpoint1()

    slash.hooks.session_start()
    assert checkpoint1.timestamp < checkpoint2.timestamp


def _maybe_decorate(decorator, flag):

    def returned(func):
        if flag:
            func = decorator(func)
        return func
    return returned


def autoname(plugin):
    def get_name(self):
        return type(self).__name__.lower()
    plugin.get_name = get_name
    return plugin



import itertools

from .._compat import itervalues

_NOTHING = object()
_END = object()


class PeekableIterator(object):

    """An iterator wrapper which allows peeking into the next element"""

    def __init__(self, iterator):
        super(PeekableIterator, self).__init__()
        self._iterator = iter(iterator)
        self._peeked = _NOTHING

    def __iter__(self):
        return self

    def next(self):
        if self._peeked is not _NOTHING:
            returned = self._peeked
            self._peeked = _NOTHING
            if returned is _END:
                raise StopIteration()
            return returned
        return next(self._iterator)
    __next__ = next

    def peek_or_none(self):
        if self.has_next():
            return self.peek()
        return None

    def peek(self):
        if self._peeked is _NOTHING:
            self._peeked = next(self._iterator, _END)
        if self._peeked is _END:
            raise StopIteration()
        return self._peeked

    def has_next(self):
        try:
            self.peek()
        except StopIteration:
            return False
        return True


def iteration(iterable):
    try:
        last_counter0 = len(iterable) - 1
    except (ValueError, TypeError):
        last_counter0 = None

    for index, element in enumerate(iterable):

        yield Iteration(element, counter0=index, counter1=index + 1, first=(index == 0), last_counter0=last_counter0)


class Iteration(object):

    def __init__(self, element, counter0, counter1, first, last_counter0=None):
        super(Iteration, self).__init__()
        self.element = element
        self.counter0 = counter0
        self.counter1 = counter1
        self.first = first
        self.last_counter0 = last_counter0
        self.last_counter1 = self.total = None if self.last_counter0 is None else self.last_counter0 + 1

    def __iter__(self):
        return iter((self, self.element))

    @property
    def last(self):
        if self.last_counter0 is None:
            raise NotImplementedError("Iterator does not support getting size")
        return self.counter0 == self.last_counter0


def iter_cartesian_dicts(d):
    """Given a dictionary of the form {name: values}, yields dictionaries corresponding to the cartesian
    product of the values, assigned to their respective names"""
    keys = list(d)  # save keys order to prevent dictionary order changes
    for combination in itertools.product(*itervalues(d)):
        yield dict(zip(keys, combination))



class Activatable(object):
    def activate(self):
        raise NotImplementedError() # pragma: no cover
    def deactivate(self):
        raise NotImplementedError() # pragma: no cover
    def __enter__(self):
        self.activate()
        return self
    def __exit__(self, *_, **__):
        self.deactivate()




from uuid import uuid4

import pytest


def test_unified_session_log_has_test_log(unified_log, log_marker):
    with open(unified_log) as f:
        assert log_marker in f.read()

def test_unified_session_log_failure_log_still_in_test_log(unified_run, test_result, log_marker):
    with open(test_result.get_log_path()) as f:
        assert log_marker in f.read()

def test_warnings_propagation(unified_run):
    assert len(unified_run.session.warnings) > 0


@pytest.fixture
def unified_log(unified_run, tmpdir):
    return str(tmpdir.join('session.log'))

@pytest.fixture
def test_result(unified_run, suite_test):
    [returned] = unified_run.get_all_results_for_test(suite_test)
    return returned


@pytest.fixture
def unified_run(suite, suite_test, tmpdir, config_override, log_marker):
    config_override('log.unified_session_log', True)
    config_override('log.root', str(tmpdir))
    config_override('log.session_subpath', 'session.log')
    suite_test.append_line('slash.logger.info({0!r})'.format(log_marker))
    suite_test.append_line('slash.logger.warning({0!r})'.format('warning'))
    return suite.run()


@pytest.fixture
def log_marker():
    return 'log_marker_{0}'.format(uuid4())



import inspect

import pytest
from slash.reporting.console_reporter import ConsoleReporter
from slash.reporting.null_reporter import NullReporter
from slash.reporting.reporter_interface import ReporterInterface
from slash.utils.python import get_underlying_func


@pytest.fixture(params=[NullReporter, ConsoleReporter])
def reporter_class(request):
    return request.param

def test_reporters_inherit_from_interface(reporter_class):
    assert issubclass(reporter_class, ReporterInterface)

def test_reporters_conform_to_interface(reporter_class):
    def _dir(cls):
        return set(name for name in dir(cls) if not name.startswith("_"))

    assert _dir(reporter_class) <= _dir(ReporterInterface)

def test_parameter_lists_conform_to_interface(reporter_class):
    for method_name in dir(reporter_class):
        if method_name.startswith("_"):
            continue

        derived_method = _get_method(reporter_class, method_name)
        base_method = _get_method(ReporterInterface, method_name)

        argpsec = inspect.getargspec(derived_method)
        expected = inspect.getargspec(base_method)
        assert argpsec == expected

def _get_method(cls, method_name):
    returned = getattr(cls, method_name)
    return get_underlying_func(returned)



import datetime
import socket
import sys
from ..interface import PluginInterface
from ...ctx import context
from ...utils.traceback_utils import get_traceback_string
from ...utils.conf_utils import Cmdline
from slash import config as slash_config
from xml.etree.ElementTree import (
    tostring as xml_to_string,
    Element as E,
    )

class Plugin(PluginInterface):

    _xunit_elements = None

    def _get_xunit_elements_list(self):
        returned = self._xunit_elements
        if returned is None:
            returned = self._xunit_elements = []
        return returned

    def get_name(self):
        return "xunit"

    def get_config(self):
        return {"filename": "testsuite.xml" // Cmdline(arg="--xunit-filename")}

    def session_start(self):
        self._start_time = datetime.datetime.now()

    def test_start(self):
        self._get_xunit_elements_list().append(E("testcase", {
            "name": str(context.test),
            "classname": type(context.test).__name__,
            "time": "0"
        }))

    def test_success(self):
        pass

    def error_added(self, result, error):
        if error.is_failure():
            self._add_error("failure")
        else:
            self._add_error("error")

    def _add_error(self, errortype):
        exc_type, exc_value, exc_tb = exc_info = sys.exc_info()
        test_element = self._get_xunit_elements_list()[-1]
        error_element = E(errortype, dict(type=exc_type.__name__, message=str(exc_value)))
        error_element.text = get_traceback_string(exc_info)
        test_element.append(error_element)

    def _get_test_case_element(self, test):
        return E('testcase', dict(name=str(test), classname="{}.{}".format(test.__class__.__module__, test.__class__.__name__), time="0"))

    def test_skip(self, reason):
        test_element = self._get_xunit_elements_list()[-1]
        test_element.append(E('skipped', type=reason))

    def result_summary(self):
        e = E('testsuite', {
            "name": "slash-suite",
            "hostname": socket.getfqdn(),
            "timestamp": self._start_time.isoformat().rsplit(".", 1)[0],
            "time": "0",
            "tests": str(context.session.results.get_num_results()),
            "errors": str(context.session.results.get_num_errors()),
            "failures": str(context.session.results.get_num_failures()),
            "skipped": str(context.session.results.get_num_skipped()),
        })
        for element in self._get_xunit_elements_list():
            e.append(element)
        with open(slash_config.root.plugin_config.xunit.filename, "wb") as outfile:
            outfile.write(xml_to_string(e))



import py.code

from slash._compat import StringIO
from .element import Element
from ..code_formatter import CodeFormatter

from contextlib import contextmanager


class CodeElement(Element):

    def __init__(self, suite):
        super(CodeElement, self).__init__(suite)
        self.suite = suite
        self._body = []

    def append_line(self, line):
        self._body.append(line)

    def prepend_line(self, line):
        self._body.insert(0, line)

    def write(self, code_formatter):
        with self._body_context(code_formatter):
            self._write_body(code_formatter)

    def append_body(self, code_element):
        """An easier way to write multiline injected code:

        @code_element.append_body
        def __code__():
            some_code_line()
            for i in range(20):
                some_other_code()
        """
        source_lines = str(py.code.Code(code_element).source()).splitlines()
        assert source_lines[0].startswith('@')
        assert source_lines[1].startswith('def ')
        assert source_lines[2][0].isspace()
        for line in str(py.code.Source('\n'.join(source_lines[2:])).deindent()).splitlines():
            self._body.append(line)

    include = append_body

    @contextmanager
    def _body_context(self, code_formatter):
        yield

    def _write_body(self, code_formatter):
        for line in self._body:
            code_formatter.writeln(line)

    @property
    def source(self):
        buff = StringIO()
        f = CodeFormatter(buff)
        self.write(f)
        return buff.getvalue()



import itertools
import sys
import time
import uuid
from contextlib import contextmanager

from .. import ctx, hooks, log
from .cleanup_manager import CleanupManager
from ..exception_handling import handling_exceptions
from ..interfaces import Activatable
from ..reporting.null_reporter import NullReporter
from ..utils.id_space import IDSpace
from ..utils.interactive import notify_if_slow_context
from ..warnings import SessionWarnings
from .fixtures.fixture_store import FixtureStore
from .result import SessionResults
from .scope_manager import ScopeManager


class Session(Activatable):
    """ Represents a slash session
    """

    duration = start_time = end_time = None

    def __init__(self, reporter=None, console_stream=None):
        super(Session, self).__init__()
        self.id = "{0}_0".format(uuid.uuid1())
        self.id_space = IDSpace(self.id)
        self.test_index_counter = itertools.count()
        self.scope_manager = ScopeManager(self)
        self._started = False
        self._complete = False
        self._active_context = None
        self.fixture_store = FixtureStore()
        self.warnings = SessionWarnings()
        self.logging = log.SessionLogging(self, console_stream=console_stream)
        #: an aggregate result summing all test results and the global result
        self.results = SessionResults(self)
        if reporter is None:
            reporter = NullReporter()
        self.reporter = reporter
        self.cleanups = CleanupManager()

    @property
    def started(self):
        return self._started

    def activate(self):
        with handling_exceptions():
            ctx.push_context()
            assert ctx.context.session is None
            assert ctx.context.result is None
            ctx.context.session = self
            ctx.context.result = self.results.global_result
            self._logging_context = self.logging.get_session_logging_context()
            self._logging_context.__enter__()
            self.cleanups.push_scope('session-global')

    def deactivate(self):
        self.results.global_result.mark_finished()
        with handling_exceptions():
            self.cleanups.pop_scope('session-global')
            self._logging_context.__exit__(*sys.exc_info())
            self._logging_context = None
            ctx.pop_context()

    @contextmanager
    def get_started_context(self):
        self.start_time = time.time()
        self.results.global_result.mark_started()
        try:
            with handling_exceptions():
                with notify_if_slow_context("Initializing session..."):
                    hooks.before_session_start()  # pylint: disable=no-member
                    hooks.session_start()  # pylint: disable=no-member
                    hooks.after_session_start()  # pylint: disable=no-member
            self._started = True
            yield
        finally:
            self._started = False
            self.results.global_result.mark_finished()
            self.end_time = time.time()
            self.duration = self.end_time - self.start_time

            with handling_exceptions():
                hooks.session_end()  # pylint: disable=no-member
            self.reporter.report_session_end(self)

    def mark_complete(self):
        self._complete = True

    def is_complete(self):
        return self._complete

    _total_num_tests = 0

    def get_total_num_tests(self):
        """Returns the total number of tests expected to run in this session
        """
        return self._total_num_tests

    def increment_total_num_tests(self, increment):
        self._total_num_tests += increment



from .element import Element


class TestContainer(Element):

    def __init__(self, suite):
        super(TestContainer, self).__init__(suite)
        self._tests = []

    @property
    def tests(self):
        return list(self._tests)

    def get_num_tests(self):
        return len(self._tests)



from slash.frontend.list_config import list_config
from slash._compat import StringIO


def test_slash_list():
    report_stream = StringIO()
    list_config([], report_stream)
    assert report_stream.getvalue()





from slash._compat import ExitStack
import slash
from slash import plugins
from slash.plugins import PluginInterface
from slash import hooks
import pytest
import gossip

from .utils import TestCase, make_runnable_tests


class SessionEndException(Exception):
    pass


class SessionStartException(Exception):
    pass


class TestEndException(Exception):
    pass


class BeforeTestCleanupException(Exception):
    pass


def test_test_skip_hook(suite, suite_test, checkpoint):
    slash.hooks.test_skip.register(checkpoint)

    suite_test.when_run.skip()

    suite.run()
    assert checkpoint.called_count == 1


@pytest.mark.parametrize('autouse', [True, False])
def test_test_start_before_fixture_start(suite, suite_test, defined_fixture, autouse):
    if autouse:
        assert not defined_fixture.autouse # make sure we're not obsolete code and that that's still where it is
        defined_fixture.autouse = True
    else:
        suite_test.depend_on_fixture(defined_fixture)
    event_code = suite.slashconf.add_hook_event('test_start', extra_args=['slash.context.test.__slash__.id'])

    summary = suite.run()
    [result] = summary.get_all_results_for_test(suite_test)
    test_id = result.test_metadata.id

    event = summary.events[event_code, test_id]

    assert summary.events['fixture_start', defined_fixture.id].timestamp > event.timestamp


def test_no_error_hooks_called_on_success(suite):

    called = []

    for hook_name in ['test_error', 'test_failure', 'test_skip', 'error_added']:
        gossip.register(lambda name=hook_name, **kw: called.append(name), 'slash.{0}'.format(hook_name))

    suite.run()
    assert not called


def test_hook__error_added_during_test(suite, request, checkpoint, suite_test):

    request.addfinalizer(
        hooks.error_added.register(checkpoint)
        .unregister)

    suite_test.when_run.raise_exception()

    summary = suite.run()
    assert checkpoint.called
    [result] = summary.get_all_results_for_test(suite_test)
    assert checkpoint.kwargs['result'] is result


def test_hook__error_added_after_test(suite, request, checkpoint, suite_test):

    request.addfinalizer(
        hooks.error_added.register(checkpoint)
        .unregister)

    summary = suite.run()
    assert not checkpoint.called
    [result] = summary.get_all_results_for_test(suite_test)
    try:
        1 / 0
    except:
        result.add_error()
    assert checkpoint.called
    assert checkpoint.kwargs['result'] is result
    assert 'ZeroDivisionError' in str(checkpoint.kwargs['error'])


def test_hook__test_interrupt(suite, request, checkpoint):
    request.addfinalizer(
        hooks.test_interrupt.register(checkpoint)
        .unregister)

    test_index = int(len(suite) / 2)
    for index, test in enumerate(suite):
        if index == test_index:
            test.when_run.interrupt()
        elif index > test_index:
            test.expect_deselect()
    suite.run(expect_interruption=True)
    assert checkpoint.called


def test_hook__test_failure_without_exception(suite, request, checkpoint, suite_test):
    request.addfinalizer(
        hooks.test_failure.register(checkpoint)
        .unregister)

    suite_test.append_line('slash.add_failure("failure")')
    suite_test.expect_failure()

    suite.run()
    assert checkpoint.called


@pytest.mark.parametrize(
    'hook_exception', [
        ('slash.session_start', SessionStartException, True),
        ('slash.session_end', SessionEndException, True),
        ('slash.test_end', TestEndException, True),
        ('slash.before_test_cleanups', BeforeTestCleanupException, False)])
@pytest.mark.parametrize('debug_enabled', [True, False])
def test_debugger_called_on_hooks(hook_exception, request, forge, config_override, checkpoint, debug_enabled):
    hook_name, exception_type, should_raise = hook_exception

    @gossip.register(hook_name)
    def raise_exc():
        raise exception_type()

    config_override("debug.enabled", debug_enabled)

    def test_something():
        pass

    forge.replace_with(slash.utils.debug, 'launch_debugger', checkpoint)

    with ExitStack() as exception_stack:
        if should_raise:
            exception_stack.enter_context(pytest.raises(exception_type))
        with slash.Session() as s:
            with s.get_started_context():
                slash.runner.run_tests(make_runnable_tests(test_something))

    assert checkpoint.called == debug_enabled
    if debug_enabled:
        assert checkpoint.args[0][0] is exception_type
        assert type(checkpoint.args[0][1]) is exception_type


def test_before_cleanup_hook(request, forge):
    cleanup = forge.create_wildcard_function_stub(name='cleanup')
    before_cleanup_hook = forge.create_wildcard_function_stub(name='before_test_cleanup')
    test_end_hook = forge.create_wildcard_function_stub(name='test_end')
    gossip.register(before_cleanup_hook, 'slash.before_test_cleanups')
    gossip.register(test_end_hook, 'slash.test_end')

    before_cleanup_hook()
    cleanup()
    test_end_hook()

    forge.replay()

    def test_something():
        slash.add_cleanup(cleanup)

    with slash.Session() as s:
        with s.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))


#### Older tests below, need modernizing ####

class HookCallingTest(TestCase):

    def setUp(self):
        super(HookCallingTest, self).setUp()
        self.plugin1 = make_custom_plugin("plugin1", self)
        self.plugin2 = make_custom_plugin("plugin2", self, hook_names=["before_session_start", "session_start", "after_session_start"])
        self.addCleanup(plugins.manager.uninstall, self.plugin1)
        self.addCleanup(plugins.manager.uninstall, self.plugin2)

    def test_hook_calling_order(self):
        # expect:
        with self.forge.any_order():
            self.plugin1.activate()
            self.plugin2.activate()

        with self.forge.any_order():
            self.plugin1.before_session_start()
            self.plugin2.before_session_start()

        with self.forge.any_order():
            self.plugin1.session_start()
            self.plugin2.session_start()

        with self.forge.any_order():
            self.plugin1.after_session_start()
            self.plugin2.after_session_start()

        self.plugin1.session_end()

        self.forge.replay()
        # get:

        plugins.manager.install(self.plugin1, activate=True)
        plugins.manager.install(self.plugin2, activate=True)

        with slash.Session() as s:
            with s.get_started_context():
                pass


def make_custom_plugin(name, test, hook_names=None):

    class CustomPlugin(PluginInterface):

        def get_name(self):
            return name

    CustomPlugin.__name__ = name

    if hook_names is None:
        hook_names = [name for name, _ in slash.hooks.get_all_hooks()]

    for hook_name in hook_names:
        setattr(CustomPlugin, hook_name, test.forge.create_wildcard_function_stub(name=hook_name))

    setattr(CustomPlugin, "activate", test.forge.create_wildcard_function_stub(name="activate"))

    return CustomPlugin()



import os

from slash import plugins

import pytest



def test_plugin_discovery(no_plugins, root_path, expected_names, config_override):
    config_override("plugins.search_paths", [root_path])
    plugins.manager.discover()
    assert set(plugins.manager.get_installed_plugins().keys()) == expected_names


@pytest.fixture
def root_path(tmpdir):
    return str(tmpdir.join('root_path'))

@pytest.fixture
def expected_names(root_path):
    returned = set()
    for index, path in enumerate([
            "a/b/p1.py",
            "a/b/p2.py",
            "a/p3.py",
            "a/b/c/p4.py",
    ]):
        plugin_name = "auto_plugin_{0}".format(index)
        path = os.path.join(root_path, path)
        if not os.path.isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        with open(path, "w") as f:
            f.write("""
import slash.plugins
from slash.plugins.interface import PluginInterface

class {name}(PluginInterface):
    def get_name(self):
        return {name!r}

def install_plugins():
""".format(name=plugin_name))
            if index % 2 == 0:
                # don't install
                f.write("     pass")
            else:
                returned.add(plugin_name)
                f.write("     slash.plugins.manager.install({name}())".format(name=plugin_name))
    for junk_file in [
            "a/junk1.p",
            "a/b/junk2",
            "a/b/c/junk3",
    ]:
        with open(os.path.join(root_path, junk_file), "w") as f:
            f.write("---JUNK----")
    return returned



from contextlib import contextmanager

from .test import Test


class NonMethodTest(Test):

    cls = None

    def __init__(self, suite, file):
        super(NonMethodTest, self).__init__(suite, file)

    def is_method_test(self):
        return False



import itertools

import pytest
import slash
from slash._compat import ExitStack, xrange

from .utils import run_tests_assert_success, make_runnable_tests
from .utils.code_formatter import CodeFormatter


def test_fixtures_representation_strings(results, a_values, fixture_values, filename, is_class):
    prefix = '{0}:'.format(filename)
    if is_class:
        prefix += 'Test.'
    assert len(results) == len(a_values) * len(fixture_values)
    assert set(result.test_metadata.address for result in results) == set(
        '{0}test_1(a={1}, fixture=fixture{2})'.format(prefix, i, j) for i, j in itertools.product(a_values, xrange(len(fixture_values))))


@pytest.mark.parametrize('non_printable', ['string/with/slashes', object()])
def test_fixtures_avoid_non_printable_reprs_strs(non_printable):
    with slash.Session():

        @slash.parametrize('param', [non_printable])
        def test_something(param):
            pass

        [loaded_test] = make_runnable_tests([test_something])

    assert '/' not in loaded_test.__slash__.address_in_file
    assert repr(non_printable) not in loaded_test.__slash__.address_in_file
    assert str(non_printable) not in loaded_test.__slash__.address_in_file


@pytest.fixture
def results(filename):

    with slash.Session() as s:
        tests = slash.loader.Loader().get_runnables(filename)
        session = run_tests_assert_success(tests, session=s)
    return list(session.results.iter_test_results())


@pytest.fixture
def filename(is_class, a_values, fixture_values, tmpdir):
    returned = str(tmpdir.join('testfile.py'))

    with open(returned, 'w') as f:
        with ExitStack() as stack:
            code = CodeFormatter(f)

            code.writeln('import slash')
            code.writeln('@slash.fixture')
            code.writeln(
                '@slash.parametrize("value", {0})'.format(fixture_values))
            code.writeln('def fixture(value):')
            with code.indented():
                code.writeln('return value')

            if is_class:
                code.writeln('class Test(slash.Test):')
                stack.enter_context(code.indented())

            code.writeln('@slash.parametrize("a", {0})'.format(a_values))
            code.write('def test_1(')
            if is_class:
                code.write('self, ')
            code.writeln('a, fixture):')
            with code.indented():
                code.writeln('pass')
    return returned


@pytest.fixture
def a_values():
    return [1, 2, 3]


@pytest.fixture
def fixture_values():
    return [4, 5, 6]


@pytest.fixture(params=[True, False])
def is_class(request):
    return request.param



import itertools
from contextlib import contextmanager

from .function import Function


class Fixture(Function):

    def __init__(self, suite, file, scope=None, autouse=False):
        super(Fixture, self).__init__(suite)
        self.file = file
        self.scope = scope
        self.autouse = autouse

    def add_cleanup(self, **kwargs):
        return self.add_deferred_event('this.add_cleanup', name='fixture_cleanup', **kwargs)

    def _write_decorators(self, code_formatter):
        self._write_fixture_decorator(code_formatter)
        super(Fixture, self)._write_decorators(code_formatter)

    def _get_function_name(self):
        return 'fx_{0}'.format(self.id)

    def _write_prologue(self, code_formatter):
        if not self.suite.debug_info:
            return

        self._write_event(code_formatter, 'fixture_start')
        code_formatter.writeln(
            '__ut__.notify_fixture_start({0!r}, {1})'.format(self.id, self._get_value_string()))
        code_formatter.writeln('@this.add_cleanup')
        code_formatter.writeln('def cleanup():')
        with code_formatter.indented():
            self._write_event(code_formatter, 'fixture_end')
            code_formatter.writeln(
                '__ut__.notify_fixture_end({0!r})'.format(self.id))

        super(Fixture, self)._write_prologue(code_formatter)

    def _write_return(self, code_formatter):
        code_formatter.writeln(
            'return {0}'.format(self._get_value_string()))

    def _write_fixture_decorator(self, code_formatter):

        params = {}
        if self.scope is not None:
            params['scope'] = self.scope
        if self.autouse:
            params['autouse'] = True

        code_formatter.write('@slash.fixture')
        if params:
            code_formatter.write('({0})'.format(', '.join(
                '{0}={1!r}'.format(k, v) for k, v in params.items())))
        code_formatter.writeln()

    def _get_value_string(self):
        returned = '{{"value": {0!r}, "params": {{'.format(self.name)

        for param in itertools.chain(self._parameters, self._fixtures):
            returned += '{0!r}: {1},'.format(param.id, param.name)
        returned += '} }'
        return returned

    def _get_argument_names(self):
        return itertools.chain(['this'], super(Fixture, self)._get_argument_names())

    def __repr__(self):
        return '<Fixture {0}>'.format(self.name)



import os
import time
from tempfile import mkdtemp

import pytest
from slash import resuming
from slash.resuming import (_MAX_NUM_SAVED_SESSIONS, CannotResume,
                            get_last_resumeable_session_id, get_tests_to_resume,
                            save_resume_state)


@pytest.fixture
def patched_resume_dir(forge):
    path = mkdtemp()
    forge.replace_with(resuming, '_RESUME_DIR', path)
    return path


def test_resume_no_session():
    with pytest.raises(CannotResume):
        get_tests_to_resume("nonexisting_session")


def test_max_resumed_sessions(suite, patched_resume_dir):
    for i in range(1, _MAX_NUM_SAVED_SESSIONS + 2):
        summary = suite.run()
        assert len(os.listdir(patched_resume_dir)) == min(_MAX_NUM_SAVED_SESSIONS, i)


def test_get_last_resumeable_session(suite):
    suite[len(suite) // 2].when_run.fail()
    prev_id = None
    for i in range(5):
        results = suite.run()
        assert results.session.id != prev_id
        prev_id = results.session.id
        save_resume_state(results.session.results)
        assert get_last_resumeable_session_id() == results.session.id


def test_resume(suite):
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    for index, test in enumerate(suite):
        if index > fail_index:
            test.expect_not_run()
    result = suite.run(additional_args=['-x'])

    save_resume_state(result.session.results)

    resumed = get_tests_to_resume(result.session.id)

    assert len(resumed) + result.session.results.get_num_started() - 1 == len(suite)
    assert resumed[0].endswith(suite[fail_index].id)



class FixtureBase(object):

    info = None
    fixture_kwargs = None
    parametrization_ids = None
    names = None

    def __init__(self):
        super(FixtureBase, self).__init__()

    def get_value(self, kwargs, active_fixture):
        raise NotImplementedError()  # pragma: no cover

    def get_variations(self):
        return None  # pragma: no cover

    def resolve(self, store):
        if self.fixture_kwargs is None:
            self.fixture_kwargs = self._resolve(store)

    def _resolve(self, store):
        raise NotImplementedError()  # pragma: no cover



import functools
import itertools

from ..._compat import izip, iteritems
from ...utils.python import getargspec
from ...utils.function_marker import function_marker

_id_gen = itertools.count(1000)

def fixture(func=None, name=None, scope=None, autouse=False):
    if func is None:
        return functools.partial(fixture, name=name, scope=scope, autouse=autouse)

    if not hasattr(func, '__slash_fixture__'):
        func.__slash_fixture__ = FixtureInfo(func, name=name, scope=scope, autouse=autouse)

    return func

nofixtures = function_marker('__slash_nofixtures__')
nofixtures.__doc__ = 'Marks the decorated function as opting out of automatic fixture deduction. Slash will not attempt to parse needed fixtures from its argument list'


class FixtureInfo(object):

    def __init__(self, func=None, name=None, scope=None, autouse=False):
        super(FixtureInfo, self).__init__()
        self.id = next(_id_gen)
        if name is None:
            if func is None:
                name = '__unnamed_{0}'.format(self.id)
            else:
                name = func.__name__
        if scope is None:
            scope = 'test'
        self.name = name
        self.func = func
        self.autouse = autouse
        self.scope = _SCOPES[scope]
        if self.func is not None:
            self.required_args = getargspec(func).args
        else:
            self.required_args = []
        if 'this' in self.required_args:
            self.required_args.remove('this')
            self.needs_this = True
        else:
            self.needs_this = False

def get_scope_by_name(scope_name):
    return _SCOPES[scope_name]

def get_scope_name_by_scope(scope_id):
    return _SCOPES_BY_ID[scope_id]

_SCOPES = dict(
    izip(('test', 'module', 'session'), itertools.count()))

_SCOPES_BY_ID = dict((id, name) for (name, id) in iteritems(_SCOPES))
